cimport croaring
from libc.stdint cimport uint32_t, uint64_t, uintptr_t, int64_t
from libcpp cimport bool
from libcpp.vector cimport vector
from libc.stdlib cimport free, malloc

from cpython cimport array
import array

try:
    range = xrange
except NameError: # python 3
    pass


cdef croaring.roaring_bitmap_t *deserialize_ptr(const unsigned char[:] buff):
    cdef croaring.roaring_bitmap_t *ptr
    cdef const char *reason_failure = NULL

    cdef char* buffer_ptr = <char*>&buff[0]

    buff_size = len(buff)
    ptr = croaring.roaring_bitmap_portable_deserialize_safe(buffer_ptr, buff_size)

    if ptr == NULL:
      raise ValueError("Could not deserialize bitmap")
    # Validate the bitmap
    if not croaring.roaring_bitmap_internal_validate(ptr, &reason_failure):
        # If validation fails, free the bitmap and raise an exception
        croaring.roaring_bitmap_free(ptr)
        raise ValueError(f"Invalid bitmap after deserialization: {reason_failure.decode('utf-8')}")
    return ptr

cdef croaring.roaring64_bitmap_t *deserialize64_ptr(const unsigned char[:] buff):
    cdef croaring.roaring64_bitmap_t *ptr
    cdef const char *reason_failure = NULL

    cdef char* buffer_ptr = <char*>&buff[0]

    buff_size = len(buff)
    ptr = croaring.roaring64_bitmap_portable_deserialize_safe(buffer_ptr, buff_size)
    if ptr == NULL:
      raise ValueError("Could not deserialize bitmap")
    # Validate the bitmap
    if not croaring.roaring64_bitmap_internal_validate(ptr, &reason_failure):
        # If validation fails, free the bitmap and raise an exception
        croaring.roaring64_bitmap_free(ptr)
        raise ValueError(f"Invalid bitmap after deserialization: {reason_failure.decode('utf-8')}")
    return ptr

def _string_rep(bm):
    skip_rows = len(bm) > 500 #this is the cutoff number for the truncating to kick in.
    table_max_width = 80  # this isn't the length of the entire output, it's only for the numeric part
    num_lines_if_skipping = 5  # the number of lines to show in the beginning and the end when output is being truncated

    head = bm.__class__.__name__ + '(['
    row_start_buffer = ' ' * len(head)
    tail = '])'

    try:
        maxval = bm.max()
    except ValueError:
        # empty bitmap
        return head + tail

    element_max_length = len(str(maxval))
    column_width = element_max_length + 2

    num_columns = table_max_width // column_width

    num_rows = (len(bm) + num_columns - 1) // num_columns
    rows = []
    row_idx = 0
    skipped = False
    while row_idx < num_rows:
        row_ints = bm[row_idx * num_columns:(row_idx + 1) * num_columns]

        line = []
        for i in row_ints:
            s = str(i)
            if num_rows == 1:
                # no padding if all numbers fit on a single line
                line.append(s)
            else:
                line.append(' ' * (element_max_length - len(s)) + s)

        if row_idx == 0:
            prefix = head
        else:
            prefix = row_start_buffer
        rows.append(prefix + ', '.join(line) + ',')
        row_idx += 1
        if skip_rows and not skipped and row_idx >= num_lines_if_skipping:
            rows.append((' ' * ((table_max_width + len(head)) // 2)) + '...')
            skipped = True
            row_idx = num_rows - num_lines_if_skipping

    rows[-1] = rows[-1].rstrip(',')  # remove trailing comma from the last line
    return '\n'.join(rows) + tail

cdef class AbstractBitMap:
    """
    An efficient and light-weight ordered set of 32 bits integers.
    """
    cdef croaring.roaring_bitmap_t* _c_bitmap
    cdef int64_t _h_val

    def __cinit__(self, values=None, copy_on_write=False, optimize=True, no_init=False):
        if no_init:
            assert values is None and not copy_on_write
            return
        cdef vector[uint32_t] buff_vect
        cdef unsigned[:] buff
        if values is None:
            self._c_bitmap = croaring.roaring_bitmap_create()
        elif isinstance(values, AbstractBitMap):
            self._c_bitmap = croaring.roaring_bitmap_copy((<AbstractBitMap?>values)._c_bitmap)
            self._h_val = (<AbstractBitMap?>values)._h_val
        elif isinstance(values, range):
            _, (start, stop, step) = values.__reduce__()
            if step < 0:
                values = range(min(values), max(values)+1, -step)
                _, (start, stop, step) = values.__reduce__()
            if start >= stop:
                self._c_bitmap = croaring.roaring_bitmap_create()
            else:
                self._c_bitmap = croaring.roaring_bitmap_from_range(start, stop, step)
        elif isinstance(values, array.array):
            size = len(values)
            if size == 0:
                self._c_bitmap = croaring.roaring_bitmap_create()
            else:
                buff = <array.array> values
                self._c_bitmap = croaring.roaring_bitmap_of_ptr(size, &buff[0])
        else:
            try:
                size = len(values)
            except TypeError:  # object has no length, creating a list
                values = list(values)
                size = len(values)
            self._c_bitmap = croaring.roaring_bitmap_create()
            if size > 0:
                buff_vect = values
                croaring.roaring_bitmap_add_many(self._c_bitmap, size, &buff_vect[0])
        if not isinstance(values, AbstractBitMap):
            croaring.roaring_bitmap_set_copy_on_write(self._c_bitmap, copy_on_write)
            self._h_val = 0
        if optimize:
            self.run_optimize()
            self.shrink_to_fit()

    def __init__(self, values=None, copy_on_write=False, optimize=True):
        """
        Construct a AbstractBitMap object, either empry or from an iterable.

        Copy on write can be enabled with the field copy_on_write.

        >>> BitMap()
        BitMap([])
        >>> BitMap([1, 123456789, 27])
        BitMap([1, 27, 123456789])
        >>> BitMap([1, 123456789, 27], copy_on_write=True)
        BitMap([1, 27, 123456789])
        """

    cdef from_ptr(self, croaring.roaring_bitmap_t *ptr) noexcept:
        """
        Return an instance of AbstractBitMap (or one of its subclasses) initialized with the given pointer.

        FIXME: this should be a classmethod, but this is (currently) impossible for cdef methods.
        See https://groups.google.com/forum/#!topic/cython-users/FLHiLzzKqj4
        """
        bm = self.__class__.__new__(self.__class__, no_init=True)
        (<AbstractBitMap>bm)._c_bitmap = ptr
        return bm

    @property
    def raw_pointer(self):
        """
        The address of the underlying `roaring_bitmap_t`, as a Python int.

        Intended for FFI interop, e.g. passing the bitmap to another
        native extension via CFFI or ctypes without a copy.

        Caveats (the caller is responsible for all of these):

        - The pointer is owned by this Python object. It is only valid
          while the object is alive; the caller must not pass it to
          `roaring_bitmap_free`.

        - Mutating the bitmap through this pointer on a `FrozenBitMap`
          is undefined behaviour: it silently invalidates the cached
          hash and breaks set/dict semantics.
        """
        return <uintptr_t>self._c_bitmap

    @property
    def copy_on_write(self):
        """
        True if and only if the bitmap has "copy on write" optimization enabled.

        >>> BitMap(copy_on_write=False).copy_on_write
        False
        >>> BitMap(copy_on_write=True).copy_on_write
        True
        """
        return croaring.roaring_bitmap_get_copy_on_write(self._c_bitmap)

    def run_optimize(self):
        return croaring.roaring_bitmap_run_optimize(self._c_bitmap)

    def shrink_to_fit(self):
        return croaring.roaring_bitmap_shrink_to_fit(self._c_bitmap)

    def __dealloc__(self):
        if self._c_bitmap is not NULL:
            croaring.roaring_bitmap_free(self._c_bitmap)

    def _check_compatibility(self, AbstractBitMap other):
        if other is None:
            raise TypeError('Argument has incorrect type (expected pyroaring.AbstractBitMap, got None)')
        if self.copy_on_write != other.copy_on_write:
            raise ValueError('Cannot have interactions between bitmaps with and without copy_on_write.\n')

    def __contains__(self, uint32_t value):
        return croaring.roaring_bitmap_contains(self._c_bitmap, value)

    def __bool__(self):
        return not croaring.roaring_bitmap_is_empty(self._c_bitmap)

    def __len__(self):
        return croaring.roaring_bitmap_get_cardinality(self._c_bitmap)

    def __lt__(self, AbstractBitMap other):
        self._check_compatibility(other)
        return croaring.roaring_bitmap_is_strict_subset((<AbstractBitMap?>self)._c_bitmap, (<AbstractBitMap?>other)._c_bitmap)

    def __le__(self, AbstractBitMap other):
        self._check_compatibility(other)
        return croaring.roaring_bitmap_is_subset((<AbstractBitMap?>self)._c_bitmap, (<AbstractBitMap?>other)._c_bitmap)

    def __eq__(self, object other):
        if not isinstance(other, AbstractBitMap):
            return NotImplemented
        self._check_compatibility(other)
        return croaring.roaring_bitmap_equals((<AbstractBitMap?>self)._c_bitmap, (<AbstractBitMap?>other)._c_bitmap)

    def __ne__(self, object other):
        if not isinstance(other, AbstractBitMap):
            return NotImplemented
        self._check_compatibility(other)
        return not croaring.roaring_bitmap_equals((<AbstractBitMap?>self)._c_bitmap, (<AbstractBitMap?>other)._c_bitmap)

    def __gt__(self, AbstractBitMap other):
        self._check_compatibility(other)
        return croaring.roaring_bitmap_is_strict_subset((<AbstractBitMap?>other)._c_bitmap, (<AbstractBitMap?>self)._c_bitmap)

    def __ge__(self, AbstractBitMap other):
        self._check_compatibility(other)
        return croaring.roaring_bitmap_is_subset((<AbstractBitMap?>other)._c_bitmap, (<AbstractBitMap?>self)._c_bitmap)

    def contains_range(self, uint64_t range_start, uint64_t range_end):
        """
        Check whether a range of values from range_start (included) to range_end (excluded) is present.

        >>> bm = BitMap([5, 6, 7, 8, 9, 10])
        >>> bm.contains_range(6, 9)
        True
        >>> bm.contains_range(8, 12)
        False
        """
        if range_end <= range_start or range_end == 0 or range_start >= 2**32:
            return True  # empty range
        if range_end >= 2**32:
            range_end = 2**32
        return croaring.roaring_bitmap_contains_range(self._c_bitmap, range_start, range_end)

    def range_cardinality(self, uint64_t range_start, uint64_t range_end):
        """
        Return cardinality from range_start (included) to range_end (excluded).

        >>> bm = BitMap(range(10))
        >>> bm.range_cardinality(0, 10)
        10
        >>> bm.range_cardinality(10, 100)
        0
        """
        if range_end < range_start:
            raise AssertionError('range_end must not be lower than range_start')
        return croaring.roaring_bitmap_range_cardinality(self._c_bitmap, range_start, range_end)

    cdef compute_hash(self):
        cdef int64_t h_val = 0
        cdef uint32_t i, count, max_count=256
        cdef croaring.roaring_uint32_iterator_t *iterator = croaring.roaring_iterator_create(self._c_bitmap)
        cdef uint32_t *buff = <uint32_t*>malloc(max_count*4)
        while True:
            count = croaring.roaring_uint32_iterator_read(iterator, buff, max_count)
            i = 0
            while i < count:
                h_val = (h_val << 2) + buff[i] + 1
                # TODO find a good hash formula
                # This one should be better, but is too long:
                # h_val = ((h_val<<16) + buff[i]) % 1748104473534059
                i += 1
            if count != max_count:
                break
        croaring.roaring_uint32_iterator_free(iterator)
        free(buff)
        if not self:
            return -1
        return h_val

    def __hash__(self):
        if self._h_val == 0: 
            self._h_val = self.compute_hash()
        return self._h_val

    def iter_equal_or_larger(self, uint32_t val):
        """
        Iterate over items in the bitmap equal or larger than a given value.

        >>> bm = BitMap([1, 2, 4])
        >>> list(bm.iter_equal_or_larger(2))
        [2, 4]
        """
        cdef croaring.roaring_uint32_iterator_t *iterator = croaring.roaring_iterator_create(self._c_bitmap)
        valid = croaring.roaring_uint32_iterator_move_equalorlarger(iterator, val)
        if not valid:
            return
        try:
            while iterator.has_value:
                yield iterator.current_value
                croaring.roaring_uint32_iterator_advance(iterator)
        finally:
            croaring.roaring_uint32_iterator_free(iterator)

    def __iter__(self):
        cdef croaring.roaring_uint32_iterator_t *iterator = croaring.roaring_iterator_create(self._c_bitmap)
        try:
            while iterator.has_value:
                yield iterator.current_value
                croaring.roaring_uint32_iterator_advance(iterator)
        finally:
            croaring.roaring_uint32_iterator_free(iterator)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return _string_rep(self)

    def flip(self, uint64_t start, uint64_t end):
        """
        Compute the negation of the bitmap within the specified interval.

        Areas outside the range are passed unchanged.

        >>> bm = BitMap([3, 12])
        >>> bm.flip(10, 15)
        BitMap([3, 10, 11, 13, 14])
        """
        return self.from_ptr(croaring.roaring_bitmap_flip(self._c_bitmap, start, end))

    def shift(self, int64_t offset):
        """
        Add the value 'offset' to each and every value of the bitmap.

        If offset + element is outside of the range [0,2^32), that the element will be dropped.

        >>> bm = BitMap([3, 12])
        >>> bm.shift(21)
        BitMap([24, 33])
        """
        return self.from_ptr(croaring.roaring_bitmap_add_offset(self._c_bitmap, offset))

    def copy(self):
        """
        Return a copy of a set.

        >>> bm = BitMap([3, 12])
        >>> bm2 = bm.copy()
        >>> bm == bm2
        True
        >>> bm.add(1)
        >>> bm == bm2
        False

        """
        return self.__class__(self)

    def isdisjoint(self, other):
        """
        Return True if two sets have a null intersection.

        >>> BitMap([1, 2]).isdisjoint(BitMap([3, 4]))
        True

        >>> BitMap([1, 2, 3]).isdisjoint(BitMap([3, 4]))
        False

        """
        return self.intersection_cardinality(other) == 0

    def issubset(self, other):
        """
        Report whether another set contains this set.

        >>> BitMap([1, 2]).issubset(BitMap([1, 2, 3, 4]))
        True

        >>> BitMap([1, 2]).issubset(BitMap([3, 4]))
        False

        """
        return self <= other

    def issuperset(self, other):
        """
        Report whether this set contains another set.

        >>> BitMap([1, 2, 3, 4]).issuperset(BitMap([1, 2]))
        True

        >>> BitMap([1, 2]).issuperset(BitMap([3, 4]))
        False

        """
        return self >= other

    def difference(*bitmaps):
        """
        Return the difference of two or more sets as a new set.

        (i.e. all elements that are in this set but not the others.)

        >>> BitMap.difference(BitMap([1, 2, 3]), BitMap([2, 20]), BitMap([3, 30]))
        BitMap([1])

        """
        size = len(bitmaps)
        cdef AbstractBitMap result, bm
        if size <= 1:
            return bitmaps[0].copy()
        elif size == 2:
            return bitmaps[0] - bitmaps[1]
        else:
            result = BitMap(bitmaps[0])
            result._h_val = 0
            for bm in bitmaps[1:]:
                result -= bm
            return bitmaps[0].__class__(result)


    def symmetric_difference(self, other):
        """
        Return the symmetric difference of two sets as a new set.

        (i.e. all elements that are in exactly one of the sets.)

        >>> BitMap([1, 2, 3]).symmetric_difference(BitMap([2, 3, 4]))
        BitMap([1, 4])
        """
        return self.__xor__(other)

    def union(*bitmaps):
        """
        Return the union of the bitmaps.

        >>> BitMap.union(BitMap([3, 12]), BitMap([5]), BitMap([0, 10, 12]))
        BitMap([0, 3, 5, 10, 12])
        """
        size = len(bitmaps)
        cdef croaring.roaring_bitmap_t *result
        cdef AbstractBitMap bm
        cdef vector[const croaring.roaring_bitmap_t*] buff
        if size <= 1:
            return bitmaps[0].copy()
        elif size == 2:
            return bitmaps[0] | bitmaps[1]
        else:
            for bm in bitmaps:
                bitmaps[0]._check_compatibility(bm)
                buff.push_back(bm._c_bitmap)
            result = croaring.roaring_bitmap_or_many(size, &buff[0])
            return (<AbstractBitMap>bitmaps[0].__class__()).from_ptr(result) # FIXME to change when from_ptr is a classmethod

    def intersection(*bitmaps): # FIXME could be more efficient
        """
        Return the intersection of the bitmaps.

        >>> BitMap.intersection(BitMap(range(0, 15)), BitMap(range(5, 20)), BitMap(range(10, 25)))
        BitMap([10, 11, 12, 13, 14])
        """
        size = len(bitmaps)
        cdef AbstractBitMap result, bm
        if size <= 1:
            return bitmaps[0].copy()
        elif size == 2:
            return bitmaps[0] & bitmaps[1]
        else:
            result = BitMap(bitmaps[0])
            result._h_val = 0
            for bm in bitmaps[1:]:
                result &= bm
            return bitmaps[0].__class__(result)

    cdef binary_op(self, AbstractBitMap other, (croaring.roaring_bitmap_t*)func(const croaring.roaring_bitmap_t*, const croaring.roaring_bitmap_t*) noexcept) noexcept:
        cdef croaring.roaring_bitmap_t *r = func(self._c_bitmap, other._c_bitmap)
        return self.from_ptr(r)

    def __or__(self, other):
        self._check_compatibility(other)
        return (<AbstractBitMap>self).binary_op(<AbstractBitMap?>other, croaring.roaring_bitmap_or)

    def __and__(self, other):
        self._check_compatibility(other)
        return (<AbstractBitMap>self).binary_op(<AbstractBitMap?>other, croaring.roaring_bitmap_and)

    def __xor__(self, other):
        self._check_compatibility(other)
        return (<AbstractBitMap>self).binary_op(<AbstractBitMap?>other, croaring.roaring_bitmap_xor)

    def __sub__(self, other):
        self._check_compatibility(other)
        return (<AbstractBitMap>self).binary_op(<AbstractBitMap?>other, croaring.roaring_bitmap_andnot)

    def union_cardinality(self, AbstractBitMap other):
        """
        Return the number of elements in the union of the two bitmaps.

        It is equivalent to len(self | other), but faster.

        >>> BitMap([3, 12]).union_cardinality(AbstractBitMap([3, 5, 8]))
        4
        """
        self._check_compatibility(other)
        return croaring.roaring_bitmap_or_cardinality(self._c_bitmap, other._c_bitmap)

    def intersection_cardinality(self, AbstractBitMap other):
        """
        Return the number of elements in the intersection of the two bitmaps.

        It is equivalent to len(self & other), but faster.

        >>> BitMap([3, 12]).intersection_cardinality(BitMap([3, 5, 8]))
        1
        """
        self._check_compatibility(other)
        return croaring.roaring_bitmap_and_cardinality(self._c_bitmap, other._c_bitmap)

    def difference_cardinality(self, AbstractBitMap other):
        """
        Return the number of elements in the difference of the two bitmaps.

        It is equivalent to len(self - other), but faster.

        >>> BitMap([3, 12]).difference_cardinality(BitMap([3, 5, 8]))
        1
        """
        self._check_compatibility(other)
        return croaring.roaring_bitmap_andnot_cardinality(self._c_bitmap, other._c_bitmap)

    def symmetric_difference_cardinality(self, AbstractBitMap other):
        """
        Return the number of elements in the symmetric difference of the two bitmaps.

        It is equivalent to len(self ^ other), but faster.

        >>> BitMap([3, 12]).symmetric_difference_cardinality(BitMap([3, 5, 8]))
        3
        """
        self._check_compatibility(other)
        return croaring.roaring_bitmap_xor_cardinality(self._c_bitmap, other._c_bitmap)

    def intersect(self, AbstractBitMap other):
        """
        Return True if and only if the two bitmaps have elements in common.

        It is equivalent to len(self & other) > 0, but faster.

        >>> BitMap([3, 12]).intersect(BitMap([3, 18]))
        True
        >>> BitMap([3, 12]).intersect(BitMap([5, 18]))
        False
        """
        self._check_compatibility(other)
        return croaring.roaring_bitmap_intersect(self._c_bitmap, other._c_bitmap)

    def jaccard_index(self, AbstractBitMap other):
        """
        Compute the Jaccard index of the two bitmaps.

        It is equivalent to len(self&other)/len(self|other), but faster.
        See https://en.wikipedia.org/wiki/Jaccard_index

        >>> BitMap([3, 10, 12]).jaccard_index(BitMap([3, 18]))
        0.25
        """
        self._check_compatibility(other)
        return croaring.roaring_bitmap_jaccard_index(self._c_bitmap, other._c_bitmap)

    def get_statistics(self):
        """
        Return relevant metrics about the bitmap.

        >>> stats = BitMap(range(18, 66000, 2)).get_statistics()
        >>> stats['cardinality']
        32991
        >>> stats['max_value']
        65998
        >>> stats['min_value']
        18
        >>> stats['n_array_containers']
        1
        >>> stats['n_bitset_containers']
        1
        >>> stats['n_bytes_array_containers']
        464
        >>> stats['n_bytes_bitset_containers']
        8192
        >>> stats['n_bytes_run_containers']
        0
        >>> stats['n_containers']
        2
        >>> stats['n_run_containers']
        0
        >>> stats['n_values_array_containers']
        232
        >>> stats['n_values_bitset_containers']
        32759
        >>> stats['n_values_run_containers']
        0
        >>> stats['sum_value'] 
        0
        """
        cdef croaring.roaring_statistics_t stat
        croaring.roaring_bitmap_statistics(self._c_bitmap, &stat)
        return stat

    def min(self):
        """
        Return the minimum element of the bitmap.

        It is equivalent to min(self), but faster.

        >>> BitMap([3, 12]).min()
        3
        """
        if len(self) == 0:
            raise ValueError('Empty roaring bitmap, there is no minimum.')
        else:
            return croaring.roaring_bitmap_minimum(self._c_bitmap)

    def max(self):
        """
        Return the maximum element of the bitmap.

        It is equivalent to max(self), but faster.

        >>> BitMap([3, 12]).max()
        12
        """
        if len(self) == 0:
            raise ValueError('Empty roaring bitmap, there is no maximum.')
        else:
            return croaring.roaring_bitmap_maximum(self._c_bitmap)

    def rank(self, uint32_t value):
        """
        Return the rank of the element in the bitmap.

        >>> BitMap([3, 12]).rank(12)
        2
        """
        return croaring.roaring_bitmap_rank(self._c_bitmap, value)

    def next_set_bit(self, uint32_t value):
        """
        Return the next set bit larger or equal to the given value.

        >>> BitMap([1, 2, 4]).next_set_bit(1)
        1

        >>> BitMap([1, 2, 4]).next_set_bit(3)
        4

        >>> BitMap([1, 2, 4]).next_set_bit(5)
        Traceback (most recent call last):
        ValueError: No value larger or equal to specified value.
        """
        try:
            return next(self.iter_equal_or_larger(value))
        except StopIteration:
            raise ValueError('No value larger or equal to specified value.')

    cdef int64_t _shift_index(self, int64_t index) except -1:
        cdef int64_t size = len(self)
        if index >= size or index < -size:
            raise IndexError('Index out of bound')
        if index < 0:
            return (index + size)
        else:
            return index

    cdef uint32_t _get_elt(self, int64_t index) except? 0:
        cdef uint64_t s_index = self._shift_index(index)
        cdef uint32_t elt
        cdef bool valid = croaring.roaring_bitmap_select(self._c_bitmap, s_index, &elt)
        if not valid:
            raise ValueError('Invalid rank')
        return elt

    cdef _get_slice(self, sl):
        """For a faster computation, different methods, depending on the slice."""
        start, stop, step = sl.indices(len(self))
        sign = 1 if step > 0 else -1
        if (sign > 0 and start >= stop) or (sign < 0 and start <= stop):  # empty chunk
            return self.__class__()
        r = range(start, stop, step)
        assert len(r) > 0
        first_elt = self._get_elt(start)
        last_elt  = self._get_elt(stop-sign)
        if abs(step) == 1:
            # Compute range size without len() to avoid OverflowError on 32-bit platforms
            # when first_elt and last_elt span a range exceeding Py_ssize_t.
            size = abs(last_elt - first_elt) + 1
            if size <= len(self) / 100:  # contiguous and small chunk of the bitmap
                return self & self.__class__(range(first_elt, last_elt+sign, step), copy_on_write=self.copy_on_write)
        # generic case
        if step < 0:
            start = r[-1]
            stop = r[0] + 1
            step = -step
        else:
            start = r[0]
            stop = r[-1] + 1
        return self._generic_get_slice(start, stop, step)

    cdef _generic_get_slice(self, uint32_t start, uint32_t stop, uint32_t step):
        """Assume that start, stop and step > 0 and that the result will not be empty."""
        cdef croaring.roaring_bitmap_t *result = croaring.roaring_bitmap_create()
        cdef croaring.roaring_uint32_iterator_t *iterator = croaring.roaring_iterator_create(self._c_bitmap)
        cdef uint32_t  count, max_count=256
        cdef uint32_t *buff = <uint32_t*>malloc(max_count*4)
        cdef uint32_t i_loc=0, i_glob=start, i_buff=0
        croaring.roaring_bitmap_set_copy_on_write(result, self.copy_on_write)
        first_elt = self._get_elt(start)
        valid = croaring.roaring_uint32_iterator_move_equalorlarger(iterator, first_elt)
        assert valid
        while True:
            count = croaring.roaring_uint32_iterator_read(iterator, buff, max_count)
            while i_buff < max_count and i_glob < stop:
                buff[i_loc] = buff[i_buff]
                i_loc += 1
                i_buff += step
                i_glob += step
            croaring.roaring_bitmap_add_many(result, i_loc, buff)
            if count != max_count or i_glob >= stop:
                break
            i_loc = 0
            i_buff = i_buff % max_count
        croaring.roaring_uint32_iterator_free(iterator)
        free(buff)
        return self.from_ptr(result)

    def __getitem__(self, value):
        if isinstance(value, int):
            return self._get_elt(value)
        elif isinstance(value, slice):
            return self._get_slice(value)
        else:
            return TypeError('Indices must be integers or slices, not %s' % type(value))

    def serialize(self):
        """
        Return the serialization of the bitmap. See AbstractBitMap.deserialize for the reverse operation.

        >>> BitMap.deserialize(BitMap([3, 12]).serialize())
        BitMap([3, 12])
        """
        cdef size_t size = croaring.roaring_bitmap_portable_size_in_bytes(self._c_bitmap)
        cdef char *buff = <char*>malloc(size)
        cdef real_size = croaring.roaring_bitmap_portable_serialize(self._c_bitmap, buff)
        result = buff[:size]
        free(buff)
        return result


    @classmethod
    def deserialize(cls, const unsigned char[:] buff):
        """
        Generate a bitmap from the given serialization. See AbstractBitMap.serialize for the reverse operation.

        >>> BitMap.deserialize(BitMap([3, 12]).serialize())
        BitMap([3, 12])
        """
        return (<AbstractBitMap>cls()).from_ptr(deserialize_ptr(buff)) # FIXME to change when from_ptr is a classmethod

    def __getstate__(self):
        return self.serialize()

    def __setstate__(self, state):
        try:                                            # compatibility between Python2 and Python3 (see #27)
            self._c_bitmap = deserialize_ptr(state)
        except TypeError:
            self._c_bitmap = deserialize_ptr(state.encode())


    def __sizeof__(self):
        cdef size_t size = croaring.roaring_bitmap_portable_size_in_bytes(self._c_bitmap)
        return size


    def to_array(self):
        """
        Return an array.array containing the elements of the bitmap, in increasing order.

        It is equivalent to array.array('I', self), but more efficient.

        >>> BitMap([3, 12]).to_array()
        array('I', [3, 12])
        """
        cdef int64_t size = len(self)
        if size == 0:
            return array.array('I', [])
        cdef array.array result = array.array('I')
        array.resize(result, size)
        cdef unsigned[:] buff = result
        croaring.roaring_bitmap_to_uint32_array(self._c_bitmap, &buff[0])
        return result


cdef class AbstractBitMap64:
    """
    An efficient and light-weight ordered set of 64 bits integers.
    """
    cdef croaring.roaring64_bitmap_t* _c_bitmap
    cdef int64_t _h_val

    def __cinit__(self, values=None, copy_on_write=False, optimize=True, no_init=False):
        if no_init:
            assert values is None
            return
        cdef vector[uint64_t] buff_vect
        cdef uint64_t[:] buff
        if values is None:
            self._c_bitmap = croaring.roaring64_bitmap_create()
        elif isinstance(values, AbstractBitMap64):
            self._c_bitmap = croaring.roaring64_bitmap_copy((<AbstractBitMap64?>values)._c_bitmap)
            self._h_val = (<AbstractBitMap64?>values)._h_val
        elif isinstance(values, range):
            _, (start, stop, step) = values.__reduce__()
            if step < 0:
                values = range(min(values), max(values)+1, -step)
                _, (start, stop, step) = values.__reduce__()
            if start >= stop:
                self._c_bitmap = croaring.roaring64_bitmap_create()
            else:
                self._c_bitmap = croaring.roaring64_bitmap_from_range(start, stop, step)
        elif isinstance(values, array.array):
            size = len(values)
            if size == 0:
                self._c_bitmap = croaring.roaring64_bitmap_create()
            else:
                buff = <array.array> values
                self._c_bitmap = croaring.roaring64_bitmap_of_ptr(size, &buff[0])
        else:
            try:
                size = len(values)
            except TypeError:  # object has no length, creating a list
                values = list(values)
                size = len(values)
            self._c_bitmap = croaring.roaring64_bitmap_create()
            if size > 0:
                buff_vect = values
                croaring.roaring64_bitmap_add_many(self._c_bitmap, size, &buff_vect[0])
        if not isinstance(values, AbstractBitMap64):
            self._h_val = 0
        if optimize:
            self.run_optimize()
            self.shrink_to_fit()

    def __init__(self, values=None, copy_on_write=False, optimize=True):
        """
        Construct a AbstractBitMap64 object, either empry or from an iterable.

        The field copy_on_write has no effect (yet).

        >>> BitMap64()
        BitMap64([])
        >>> BitMap64([1, 123456789, 27])
        BitMap64([1, 27, 123456789])
        """

    cdef from_ptr(self, croaring.roaring64_bitmap_t *ptr) noexcept:
        """
        Return an instance of AbstractBitMap64 (or one of its subclasses) initialized with the given pointer.

        FIXME: this should be a classmethod, but this is (currently) impossible for cdef methods.
        See https://groups.google.com/forum/#!topic/cython-users/FLHiLzzKqj4
        """
        bm = self.__class__.__new__(self.__class__, no_init=True)
        (<AbstractBitMap64>bm)._c_bitmap = ptr
        return bm

    @property
    def raw_pointer(self):
        """
        The address of the underlying `roaring_bitmap_t`, as a Python int.

        Intended for FFI interop, e.g. passing the bitmap to another
        native extension via CFFI or ctypes without a copy.

        Caveats (the caller is responsible for all of these):

        - The pointer is owned by this Python object. It is only valid
          while the object is alive; the caller must not pass it to
          `roaring_bitmap_free`.

        - Mutating the bitmap through this pointer on a `FrozenBitMap`
          is undefined behaviour: it silently invalidates the cached
          hash and breaks set/dict semantics.
        """
        return <uintptr_t>self._c_bitmap

    @property
    def copy_on_write(self):
        """
        Always False, not implemented for 64 bits roaring bitmaps.

        >>> BitMap64(copy_on_write=False).copy_on_write
        False
        >>> BitMap64(copy_on_write=True).copy_on_write
        False
        """
        return False

    def run_optimize(self):
        return croaring.roaring64_bitmap_run_optimize(self._c_bitmap)

    def shrink_to_fit(self):
        return croaring.roaring64_bitmap_shrink_to_fit(self._c_bitmap)

    def __dealloc__(self):
        if self._c_bitmap is not NULL:
            croaring.roaring64_bitmap_free(self._c_bitmap)

    def _check_compatibility(self, AbstractBitMap64 other):
        if other is None:
            raise TypeError('Argument has incorrect type (expected pyroaring.AbstractBitMap64, got None)')
        if self.copy_on_write != other.copy_on_write:
            raise ValueError('Cannot have interactions between bitmaps with and without copy_on_write.\n')

    def __contains__(self, uint64_t value):
        return croaring.roaring64_bitmap_contains(self._c_bitmap, value)

    def __bool__(self):
        return not croaring.roaring64_bitmap_is_empty(self._c_bitmap)

    def __len__(self):
        return croaring.roaring64_bitmap_get_cardinality(self._c_bitmap)

    def __lt__(self, AbstractBitMap64 other):
        self._check_compatibility(other)
        return croaring.roaring64_bitmap_is_strict_subset((<AbstractBitMap64?>self)._c_bitmap, (<AbstractBitMap64?>other)._c_bitmap)

    def __le__(self, AbstractBitMap64 other):
        self._check_compatibility(other)
        return croaring.roaring64_bitmap_is_subset((<AbstractBitMap64?>self)._c_bitmap, (<AbstractBitMap64?>other)._c_bitmap)

    def __eq__(self, object other):
        if not isinstance(other, AbstractBitMap64):
            return NotImplemented
        self._check_compatibility(other)
        return croaring.roaring64_bitmap_equals((<AbstractBitMap64?>self)._c_bitmap, (<AbstractBitMap64?>other)._c_bitmap)

    def __ne__(self, object other):
        if not isinstance(other, AbstractBitMap64):
            return NotImplemented
        self._check_compatibility(other)
        return not croaring.roaring64_bitmap_equals((<AbstractBitMap64?>self)._c_bitmap, (<AbstractBitMap64?>other)._c_bitmap)

    def __gt__(self, AbstractBitMap64 other):
        self._check_compatibility(other)
        return croaring.roaring64_bitmap_is_strict_subset((<AbstractBitMap64?>other)._c_bitmap, (<AbstractBitMap64?>self)._c_bitmap)

    def __ge__(self, AbstractBitMap64 other):
        self._check_compatibility(other)
        return croaring.roaring64_bitmap_is_subset((<AbstractBitMap64?>other)._c_bitmap, (<AbstractBitMap64?>self)._c_bitmap)

    def contains_range(self, uint64_t range_start, uint64_t range_end):
        """
        Check whether a range of values from range_start (included) to range_end (excluded) is present.

        >>> bm = BitMap64([5, 6, 7, 8, 9, 10])
        >>> bm.contains_range(6, 9)
        True
        >>> bm.contains_range(8, 12)
        False
        """
        if range_end <= range_start or range_end == 0:
            return True  # empty range
        return croaring.roaring64_bitmap_contains_range(self._c_bitmap, range_start, range_end)

    def range_cardinality(self, uint64_t range_start, uint64_t range_end):
        """
        Return cardinality from range_start (included) to range_end (excluded).

        >>> bm = BitMap64(range(10))
        >>> bm.range_cardinality(0, 10)
        10
        >>> bm.range_cardinality(10, 100)
        0
        """
        if range_end < range_start:
            raise AssertionError('range_end must not be lower than range_start')
        return croaring.roaring64_bitmap_range_cardinality(self._c_bitmap, range_start, range_end)

    cdef compute_hash(self):
        cdef int64_t h_val = 0
        cdef uint32_t i, count, max_count=256
        cdef croaring.roaring64_iterator_t *iterator = croaring.roaring64_iterator_create(self._c_bitmap)
        cdef uint64_t *buff = <uint64_t*>malloc(max_count*8)
        while True:
            count = croaring.roaring64_iterator_read(iterator, buff, max_count)
            i = 0
            while i < count:
                h_val += buff[i]
                # TODO find a good hash formula
                i += 1
            if count != max_count:
                break
        croaring.roaring64_iterator_free(iterator)
        free(buff)
        if not self:
            return -1
        return h_val

    def __hash__(self):
        if self._h_val == 0:
            self._h_val = self.compute_hash()
        return self._h_val

    def iter_equal_or_larger(self, uint64_t val):
        """
        Iterate over items in the bitmap equal or larger than a given value.

        >>> bm = BitMap64([1, 2, 4])
        >>> list(bm.iter_equal_or_larger(2))
        [2, 4]
        """
        cdef croaring.roaring64_iterator_t *iterator = croaring.roaring64_iterator_create(self._c_bitmap)
        valid = croaring.roaring64_iterator_move_equalorlarger(iterator, val)
        if not valid:
            return
        try:
            while valid:
                yield croaring.roaring64_iterator_value(iterator)
                valid = croaring.roaring64_iterator_advance(iterator)
        finally:
            croaring.roaring64_iterator_free(iterator)

    def __iter__(self):
        cdef croaring.roaring64_iterator_t *iterator = croaring.roaring64_iterator_create(self._c_bitmap)
        valid = croaring.roaring64_iterator_has_value(iterator)
        if not valid:
            return
        try:
            while valid:
                yield croaring.roaring64_iterator_value(iterator)
                valid = croaring.roaring64_iterator_advance(iterator)
        finally:
            croaring.roaring64_iterator_free(iterator)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return _string_rep(self)

    def flip(self, uint64_t start, uint64_t end):
        """
        Compute the negation of the bitmap within the specified interval.

        Areas outside the range are passed unchanged.

        >>> bm = BitMap64([3, 12])
        >>> bm.flip(10, 15)
        BitMap64([3, 10, 11, 13, 14])
        """
        return self.from_ptr(croaring.roaring64_bitmap_flip(self._c_bitmap, start, end))

    def shift(self, int64_t offset):
        """
        Add the value 'offset' to each and every value of the bitmap.

        If offset + element is outside of the range [0,2^64), that the element will be dropped.

        >>> bm = BitMap64([3, 12])
        >>> bm.shift(21)
        BitMap64([24, 33])
        """
        if offset >= 0:
            return self.from_ptr(croaring.roaring64_bitmap_add_offset_signed(self._c_bitmap, True, <uint64_t>offset))
        else:
            return self.from_ptr(croaring.roaring64_bitmap_add_offset_signed(self._c_bitmap, False, <uint64_t>0 - <uint64_t>offset))

    def get_statistics(self):
        """
        Return relevant metrics about the bitmap.

        >>> stats = BitMap64(range(18, 66000, 2)).get_statistics()
        >>> stats['cardinality']
        32991
        >>> stats['max_value']
        65998
        >>> stats['min_value']
        18
        >>> stats['n_array_containers']
        1
        >>> stats['n_bitset_containers']
        1
        >>> stats['n_bytes_array_containers']
        464
        >>> stats['n_bytes_bitset_containers']
        8192
        >>> stats['n_bytes_run_containers']
        0
        >>> stats['n_containers']
        2
        >>> stats['n_run_containers']
        0
        >>> stats['n_values_array_containers']
        232
        >>> stats['n_values_bitset_containers']
        32759
        >>> stats['n_values_run_containers']
        0
        """
        cdef croaring.roaring64_statistics_t stat
        croaring.roaring64_bitmap_statistics(self._c_bitmap, &stat)
        return stat

    def min(self):
        """
        Return the minimum element of the bitmap.

        It is equivalent to min(self), but faster.

        >>> BitMap64([3, 12]).min()
        3
        """
        if len(self) == 0:
            raise ValueError('Empty roaring bitmap, there is no minimum.')
        else:
            return croaring.roaring64_bitmap_minimum(self._c_bitmap)

    def max(self):
        """
        Return the maximum element of the bitmap.

        It is equivalent to max(self), but faster.

        >>> BitMap64([3, 12]).max()
        12
        """
        if len(self) == 0:
            raise ValueError('Empty roaring bitmap, there is no maximum.')
        else:
            return croaring.roaring64_bitmap_maximum(self._c_bitmap)

    def rank(self, uint64_t value):
        """
        Return the rank of the element in the bitmap.

        >>> BitMap64([3, 12]).rank(12)
        2
        """
        return croaring.roaring64_bitmap_rank(self._c_bitmap, value)

    def next_set_bit(self, uint64_t value):
        """
        Return the next set bit larger or equal to the given value.

        >>> BitMap64([1, 2, 4]).next_set_bit(1)
        1

        >>> BitMap64([1, 2, 4]).next_set_bit(3)
        4

        >>> BitMap64([1, 2, 4]).next_set_bit(5)
        Traceback (most recent call last):
        ValueError: No value larger or equal to specified value.
        """
        try:
            return next(self.iter_equal_or_larger(value))
        except StopIteration:
            raise ValueError('No value larger or equal to specified value.')

    cdef int64_t _shift_index(self, int64_t index) except -1:
        cdef int64_t size = len(self)
        if index >= size or index < -size:
            raise IndexError('Index out of bound')
        if index < 0:
            return (index + size)
        else:
            return index

    cdef uint64_t _get_elt(self, int64_t index) except? 0:
        cdef uint64_t s_index = self._shift_index(index)
        cdef uint64_t elt
        cdef bool valid = croaring.roaring64_bitmap_select(self._c_bitmap, s_index, &elt)
        if not valid:
            raise ValueError('Invalid rank')
        return elt

    cdef _get_slice(self, sl):
        """For a faster computation, different methods, depending on the slice."""
        start, stop, step = sl.indices(len(self))
        sign = 1 if step > 0 else -1
        if (sign > 0 and start >= stop) or (sign < 0 and start <= stop):  # empty chunk
            return self.__class__()
        r = range(start, stop, step)
        assert len(r) > 0
        first_elt = self._get_elt(start)
        last_elt  = self._get_elt(stop-sign)
        if abs(step) == 1:
            # Compute range size without len() to avoid OverflowError on platforms
            # where Py_ssize_t cannot hold the distance between first_elt and last_elt.
            size = abs(last_elt - first_elt) + 1
            if size <= len(self) / 100:  # contiguous and small chunk of the bitmap
                return self & self.__class__(range(first_elt, last_elt+sign, step))
        # generic case
        if step < 0:
            start = r[-1]
            stop = r[0] + 1
            step = -step
        else:
            start = r[0]
            stop = r[-1] + 1
        return self._generic_get_slice(start, stop, step)

    cdef _generic_get_slice(self, uint64_t start, uint64_t stop, uint64_t step):
        """Assume that start, stop and step > 0 and that the result will not be empty."""
        cdef croaring.roaring64_bitmap_t *result = croaring.roaring64_bitmap_create()
        cdef croaring.roaring64_iterator_t *iterator = croaring.roaring64_iterator_create(self._c_bitmap)
        cdef uint64_t  count, max_count=256
        cdef uint64_t *buff = <uint64_t*>malloc(max_count*8)
        cdef uint64_t i_loc=0, i_glob=start, i_buff=0
        first_elt = self._get_elt(start)
        valid = croaring.roaring64_iterator_move_equalorlarger(iterator, first_elt)
        assert valid
        while True:
            count = croaring.roaring64_iterator_read(iterator, buff, max_count)
            while i_buff < max_count and i_glob < stop:
                buff[i_loc] = buff[i_buff]
                i_loc += 1
                i_buff += step
                i_glob += step
            croaring.roaring64_bitmap_add_many(result, i_loc, buff)
            if count != max_count or i_glob >= stop:
                break
            i_loc = 0
            i_buff = i_buff % max_count
        croaring.roaring64_iterator_free(iterator)
        free(buff)
        return self.from_ptr(result)

    def __getitem__(self, value):
        if isinstance(value, int):
            return self._get_elt(value)
        elif isinstance(value, slice):
            return self._get_slice(value)
        else:
            return TypeError('Indices must be integers or slices, not %s' % type(value))

    def serialize(self):
        """
        Return the serialization of the bitmap. See AbstractBitMap64.deserialize for the reverse operation.

        >>> BitMap64.deserialize(BitMap64([3, 12]).serialize())
        BitMap64([3, 12])
        """
        cdef size_t size = croaring.roaring64_bitmap_portable_size_in_bytes(self._c_bitmap)
        cdef char *buff = <char*>malloc(size)
        cdef real_size = croaring.roaring64_bitmap_portable_serialize(self._c_bitmap, buff)
        result = buff[:size]
        free(buff)
        return result


    @classmethod
    def deserialize(cls, const unsigned char[:] buff):
        """
        Generate a bitmap from the given serialization. See AbstractBitMap64.serialize for the reverse operation.

        >>> BitMap64.deserialize(BitMap64([3, 12]).serialize())
        BitMap64([3, 12])
        """
        return (<AbstractBitMap64>cls()).from_ptr(deserialize64_ptr(buff)) # FIXME to change when from_ptr is a classmethod

    def __getstate__(self):
        return self.serialize()

    def __setstate__(self, state):
        try:                                            # compatibility between Python2 and Python3 (see #27)
            self._c_bitmap = deserialize64_ptr(state)
        except TypeError:
            self._c_bitmap = deserialize64_ptr(state.encode())


    def __sizeof__(self):
        cdef size_t size = croaring.roaring64_bitmap_portable_size_in_bytes(self._c_bitmap)
        return size

    def to_array(self):
        """
        Return an array.array containing the elements of the bitmap, in increasing order.

        It is equivalent to array.array('Q', self), but more efficient.

        >>> BitMap64([3, 12]).to_array()
        array('Q', [3, 12])
        """
        cdef uint64_t size = len(self)
        if size == 0:
            return array.array('Q', [])
        cdef array.array result = array.array('Q')
        array.resize(result, size)
        cdef uint64_t[:] buff = result
        croaring.roaring64_bitmap_to_uint64_array(self._c_bitmap, &buff[0])
        return result

    def copy(self):
        """
        Return a copy of a set.

        >>> bm = BitMap64([3, 12])
        >>> bm2 = bm.copy()
        >>> bm == bm2
        True
        >>> bm.add(1)
        >>> bm == bm2
        False

        """
        return self.__class__(self)

    def isdisjoint(self, other):
        """
        Return True if two sets have a null intersection.

        >>> BitMap64([1, 2]).isdisjoint(BitMap64([3, 4]))
        True

        >>> BitMap64([1, 2, 3]).isdisjoint(BitMap64([3, 4]))
        False

        """
        return self.intersection_cardinality(other) == 0

    def issubset(self, other):
        """
        Report whether another set contains this set.

        >>> BitMap64([1, 2]).issubset(BitMap64([1, 2, 3, 4]))
        True

        >>> BitMap64([1, 2]).issubset(BitMap64([3, 4]))
        False

        """
        return self <= other

    def issuperset(self, other):
        """
        Report whether this set contains another set.

        >>> BitMap64([1, 2, 3, 4]).issuperset(BitMap64([1, 2]))
        True

        >>> BitMap64([1, 2]).issuperset(BitMap64([3, 4]))
        False

        """
        return self >= other

    def difference(*bitmaps):
        """
        Return the difference of two or more sets as a new set.

        (i.e. all elements that are in this set but not the others.)

        >>> BitMap64.difference(BitMap64([1, 2, 3]), BitMap64([2, 20]), BitMap64([3, 30]))
        BitMap64([1])

        """
        size = len(bitmaps)
        cdef AbstractBitMap64 result, bm
        if size <= 1:
            return bitmaps[0].copy()
        elif size == 2:
            return bitmaps[0] - bitmaps[1]
        else:
            result = BitMap64(bitmaps[0])
            result._h_val = 0
            for bm in bitmaps[1:]:
                result -= bm
            return bitmaps[0].__class__(result)


    def symmetric_difference(self, other):
        """
        Return the symmetric difference of two sets as a new set.

        (i.e. all elements that are in exactly one of the sets.)

        >>> BitMap64([1, 2, 3]).symmetric_difference(BitMap64([2, 3, 4]))
        BitMap64([1, 4])
        """
        return self.__xor__(other)

    def union(*bitmaps):
        """
        Return the union of the bitmaps.

        >>> BitMap64.union(BitMap64([3, 12]), BitMap64([5]), BitMap64([0, 10, 12]))
        BitMap64([0, 3, 5, 10, 12])
        """
        size = len(bitmaps)
        cdef AbstractBitMap64 result, bm
        if size <= 1:
            return bitmaps[0].copy()
        elif size == 2:
            return bitmaps[0] | bitmaps[1]
        else:
            result = BitMap64(bitmaps[0])
            for bm in bitmaps[1:]:
                result |= bm
            return bitmaps[0].__class__(result)

    def intersection(*bitmaps):
        """
        Return the intersection of the bitmaps.

        >>> BitMap64.intersection(BitMap64(range(0, 15)), BitMap64(range(5, 20)), BitMap64(range(10, 25)))
        BitMap64([10, 11, 12, 13, 14])
        """
        size = len(bitmaps)
        cdef AbstractBitMap64 result, bm
        if size <= 1:
            return bitmaps[0].copy()
        elif size == 2:
            return bitmaps[0] & bitmaps[1]
        else:
            result = BitMap64(bitmaps[0])
            result._h_val = 0
            for bm in bitmaps[1:]:
                result &= bm
            return bitmaps[0].__class__(result)

    cdef binary_op(self, AbstractBitMap64 other, (croaring.roaring64_bitmap_t*)func(const croaring.roaring64_bitmap_t*, const croaring.roaring64_bitmap_t*) noexcept) noexcept:
        cdef croaring.roaring64_bitmap_t *r = func(self._c_bitmap, other._c_bitmap)
        return self.from_ptr(r)

    def __or__(self, other):
        return (<AbstractBitMap64>self).binary_op(<AbstractBitMap64?>other, croaring.roaring64_bitmap_or)

    def __and__(self, other):
        return (<AbstractBitMap64>self).binary_op(<AbstractBitMap64?>other, croaring.roaring64_bitmap_and)

    def __xor__(self, other):
        return (<AbstractBitMap64>self).binary_op(<AbstractBitMap64?>other, croaring.roaring64_bitmap_xor)

    def __sub__(self, other):
        return (<AbstractBitMap64>self).binary_op(<AbstractBitMap64?>other, croaring.roaring64_bitmap_andnot)

    def union_cardinality(self, AbstractBitMap64 other):
        """
        Return the number of elements in the union of the two bitmaps.

        It is equivalent to len(self | other), but faster.

        >>> BitMap64([3, 12]).union_cardinality(BitMap64([3, 5, 8]))
        4
        """
        return croaring.roaring64_bitmap_or_cardinality(self._c_bitmap, other._c_bitmap)

    def intersection_cardinality(self, AbstractBitMap64 other):
        """
        Return the number of elements in the intersection of the two bitmaps.

        It is equivalent to len(self & other), but faster.

        >>> BitMap64([3, 12]).intersection_cardinality(BitMap64([3, 5, 8]))
        1
        """
        return croaring.roaring64_bitmap_and_cardinality(self._c_bitmap, other._c_bitmap)

    def difference_cardinality(self, AbstractBitMap64 other):
        """
        Return the number of elements in the difference of the two bitmaps.

        It is equivalent to len(self - other), but faster.

        >>> BitMap64([3, 12]).difference_cardinality(BitMap64([3, 5, 8]))
        1
        """
        return croaring.roaring64_bitmap_andnot_cardinality(self._c_bitmap, other._c_bitmap)

    def symmetric_difference_cardinality(self, AbstractBitMap64 other):
        """
        Return the number of elements in the symmetric difference of the two bitmaps.

        It is equivalent to len(self ^ other), but faster.

        >>> BitMap64([3, 12]).symmetric_difference_cardinality(BitMap64([3, 5, 8]))
        3
        """
        return croaring.roaring64_bitmap_xor_cardinality(self._c_bitmap, other._c_bitmap)

    def intersect(self, AbstractBitMap64 other):
        """
        Return True if and only if the two bitmaps have elements in common.

        It is equivalent to len(self & other) > 0, but faster.

        >>> BitMap64([3, 12]).intersect(BitMap64([3, 18]))
        True
        >>> BitMap64([3, 12]).intersect(BitMap64([5, 18]))
        False
        """
        return croaring.roaring64_bitmap_intersect(self._c_bitmap, other._c_bitmap)

    def jaccard_index(self, AbstractBitMap64 other):
        """
        Compute the Jaccard index of the two bitmaps.

        It is equivalent to len(self&other)/len(self|other), but faster.
        See https://en.wikipedia.org/wiki/Jaccard_index

        >>> BitMap64([3, 10, 12]).jaccard_index(BitMap64([3, 18]))
        0.25
        """
        return croaring.roaring64_bitmap_jaccard_index(self._c_bitmap, other._c_bitmap)



from __future__ import annotations  # for using set[int] in Python 3.8

import hypothesis.strategies as st
from hypothesis.database import DirectoryBasedExampleDatabase
from hypothesis.stateful import Bundle, RuleBasedStateMachine, rule
from hypothesis import settings
from dataclasses import dataclass
from pyroaring import BitMap, BitMap64
from test import hyp_collection, uint32, uint64, is_32_bits

if is_32_bits:
    BitMapClass = BitMap
    int_class = uint32
    large_val = 2**30
else:
    BitMapClass = BitMap64
    int_class = uint64
    large_val = 2**40

@dataclass
class Collection:
    test: BitMapClass
    ref: set[int]

    def check(self):
        assert len(self.test) == len(self.ref)
        assert set(self.test) == self.ref

    def __post_init__(self):
        self.check()


class SetComparison(RuleBasedStateMachine):
    collections = Bundle("collections")

    @rule(target=collections, val=hyp_collection)
    def init_collection(self, val):
        return Collection(test=BitMapClass(val), ref=set(val))

    @rule(target=collections, col=collections)
    def copy(self, col):
        return Collection(test=BitMapClass(col.test), ref=set(col.ref))

    @rule(col=collections, val=int_class)
    def add_elt(self, col, val):
        col.test.add(val)
        col.ref.add(val)
        col.check()

    @rule(col=collections, val=int_class)
    def remove_elt(self, col, val):
        col.test.discard(val)
        col.ref.discard(val)
        col.check()

    @rule(target=collections, col1=collections, col2=collections)
    def union(self, col1, col2):
        return Collection(test=col1.test | col2.test, ref=col1.ref | col2.ref)

    @rule(col1=collections, col2=collections)
    def union_inplace(self, col1, col2):
        col1.test |= col2.test
        col1.ref |= col2.ref
        col1.check()

    @rule(target=collections, col1=collections, col2=collections)
    def intersection(self, col1, col2):
        return Collection(test=col1.test & col2.test, ref=col1.ref & col2.ref)

    @rule(col1=collections, col2=collections)
    def intersection_inplace(self, col1, col2):
        col1.test &= col2.test
        col1.ref &= col2.ref
        col1.check()

    @rule(target=collections, col1=collections, col2=collections)
    def difference(self, col1, col2):
        return Collection(test=col1.test - col2.test, ref=col1.ref - col2.ref)

    @rule(col1=collections, col2=collections)
    def difference_inplace(self, col1, col2):
        col1.test -= col2.test
        col1.ref -= col2.ref
        col1.check()

    @rule(target=collections, col1=collections, col2=collections)
    def symmetric_difference(self, col1, col2):
        return Collection(test=col1.test ^ col2.test, ref=col1.ref ^ col2.ref)

    @rule(col1=collections, col2=collections)
    def symmetric_difference_inplace(self, col1, col2):
        col1.test ^= col2.test
        col1.ref ^= col2.ref
        col1.check()

    @rule(
        target=collections,
        col=collections,
        start=st.integers(min_value=0, max_value=large_val),
        size=st.integers(min_value=0, max_value=2**18),
    )
    def flip(self, col, start, size):
        stop = start + size
        return Collection(
            test=col.test.flip(start, stop), ref=col.ref ^ set(range(start, stop))
        )

    @rule(
        col=collections,
        start=st.integers(min_value=0, max_value=large_val),
        size=st.integers(min_value=0, max_value=2**18),
    )
    def flip_inplace(self, col, start, size):
        stop = start + size
        col.test.flip_inplace(start, stop)
        col.ref ^= set(range(start, stop))
        col.check()


TestTrees = SetComparison.TestCase
TestTrees.settings = settings(max_examples=100, stateful_step_count=100)

if __name__ == "__main__":
    unittest.main()



#!python
"""
Taken from https://github.com/cython/cython/wiki/FAQ#how-can-i-run-doctests-in-cython-code-pyx-files

Cython-compatible wrapper for doctest.testmod().

Usage example, assuming a Cython module mymod.pyx is compiled.
This is run from the command line, passing a command to Python:
python -c "import cydoctest, mymod; cydoctest.testmod(mymod)"

(This still won't let a Cython module run its own doctests
when called with "python mymod.py", but it's pretty close.
Further options can be passed to testmod() as desired, e.g.
verbose=True.)
"""

import sys
import doctest
import inspect


def _from_module(module, object):
    """
    Return true if the given object is defined in the given module.
    """
    if module is None:
        return True
    elif inspect.getmodule(object) is not None:
        return module is inspect.getmodule(object)
    elif inspect.isfunction(object):
        return module.__dict__ is object.func_globals
    elif inspect.isclass(object):
        return module.__name__ == object.__module__
    elif hasattr(object, '__module__'):
        return module.__name__ == object.__module__
    elif isinstance(object, property):
        return True  # [XX] no way not be sure.
    else:
        raise ValueError("object must be a class or function")


def fix_module_doctest(module):
    """
    Extract docstrings from cython functions, that would be skipped by doctest
    otherwise.
    """
    module.__test__ = {}
    for name in dir(module):
        value = getattr(module, name)
        if inspect.isbuiltin(value) and isinstance(value.__doc__, str) and _from_module(module, value):
            module.__test__[name] = value.__doc__


def testmod(m=None, *args, **kwargs):
    """
    Fix a Cython module's doctests, then call doctest.testmod()

    All other arguments are passed directly to doctest.testmod().
    """
    fix_module_doctest(m)
    result = doctest.testmod(m, *args, **kwargs)
    if result.failed > 0:
        sys.exit('%d test(s) failed' % result.failed)


if __name__ == "__main__":
    import pyroaring
    testmod(pyroaring)



cdef class BitMap(AbstractBitMap):

    cdef compute_hash(self):
        '''Unsupported method.'''
        # For some reason, if we directly override __hash__ (either in BitMap or in FrozenBitMap), the __richcmp__
        # method disappears.
        raise TypeError('Cannot compute the hash of a %s.' % self.__class__.__name__)

    def add(self, uint32_t value):
        """
        Add an element to the bitmap. This has no effect if the element is already present.

        >>> bm = BitMap()
        >>> bm.add(42)
        >>> bm
        BitMap([42])
        >>> bm.add(42)
        >>> bm
        BitMap([42])
        """
        croaring.roaring_bitmap_add(self._c_bitmap, value)

    def add_checked(self, uint32_t value):
        """
        Add an element to the bitmap. This raises a KeyError exception if the element is already present.

        >>> bm = BitMap()
        >>> bm.add_checked(42)
        >>> bm
        BitMap([42])
        >>> bm.add_checked(42)
        Traceback (most recent call last):
        ...
        KeyError: 42
        """
        cdef bool test = croaring.roaring_bitmap_add_checked(self._c_bitmap, value)
        if not test:
            raise KeyError(value)

    def update(self, *all_values): # FIXME could be more efficient
        """
        Add all the given values to the bitmap.

        >>> bm = BitMap([3, 12])
        >>> bm.update([8, 12, 55, 18])
        >>> bm
        BitMap([3, 8, 12, 18, 55])
        """
        cdef vector[uint32_t] buff_vect
        cdef unsigned[:] buff
        for values in all_values:
            if isinstance(values, AbstractBitMap):
                self |= values
            elif isinstance(values, range):
                if len(values) == 0:
                    continue
                _, (start, stop, step) = values.__reduce__()
                if step == -1:
                    step = 1
                    start, stop = stop+1, start+1
                if step == 1:
                    self.add_range(start, stop)
                else:
                    self |= AbstractBitMap(values, copy_on_write=self.copy_on_write)
            elif isinstance(values, array.array) and len(values) > 0:
                buff = <array.array> values
                croaring.roaring_bitmap_add_many(self._c_bitmap, len(values), &buff[0])
            else:
                try:
                    size = len(values)
                except TypeError:  # object has no length, creating a list
                    values = list(values)
                    size = len(values)
                if size > 0:
                    buff_vect = values
                    croaring.roaring_bitmap_add_many(self._c_bitmap, size, &buff_vect[0])

    def discard(self, uint32_t value):
        """
        Remove an element from the bitmap. This has no effect if the element is not present.

        >>> bm = BitMap([3, 12])
        >>> bm.discard(3)
        >>> bm
        BitMap([12])
        >>> bm.discard(3)
        >>> bm
        BitMap([12])
        """
        croaring.roaring_bitmap_remove(self._c_bitmap, value)

    def remove(self, uint32_t value):
        """
        Remove an element from the bitmap. This raises a KeyError exception if the element does not exist in the bitmap.

        >>> bm = BitMap([3, 12])
        >>> bm.remove(3)
        >>> bm
        BitMap([12])
        >>> bm.remove(3)
        Traceback (most recent call last):
        ...
        KeyError: 3
        """
        cdef bool test = croaring.roaring_bitmap_remove_checked(self._c_bitmap, value)
        if not test:
            raise KeyError(value)

    cdef binary_iop(self, AbstractBitMap other, (void)func(croaring.roaring_bitmap_t*, const croaring.roaring_bitmap_t*) noexcept) noexcept:
        func(self._c_bitmap, other._c_bitmap)
        return self

    def __ior__(self, other):
        self._check_compatibility(other)
        if self._c_bitmap == (<AbstractBitMap?>other)._c_bitmap:
            return self
        return (<BitMap>self).binary_iop(<AbstractBitMap?>other, croaring.roaring_bitmap_or_inplace)

    def __iand__(self, other):
        self._check_compatibility(other)
        if self._c_bitmap == (<AbstractBitMap?>other)._c_bitmap:
            return self
        return (<BitMap>self).binary_iop(<AbstractBitMap?>other, croaring.roaring_bitmap_and_inplace)

    def __ixor__(self, other):
        self._check_compatibility(other)
        if self._c_bitmap == (<AbstractBitMap?>other)._c_bitmap:
            self.clear()
            return self
        return (<BitMap>self).binary_iop(<AbstractBitMap?>other, croaring.roaring_bitmap_xor_inplace)

    def __isub__(self, other):
        self._check_compatibility(other)
        if self._c_bitmap == (<AbstractBitMap?>other)._c_bitmap:
            self.clear()
            return self
        return (<BitMap>self).binary_iop(<AbstractBitMap?>other, croaring.roaring_bitmap_andnot_inplace)

    def intersection_update(self, *all_values): # FIXME could be more efficient
        """
        Update the bitmap by taking its intersection with the given values.

        >>> bm = BitMap([3, 12])
        >>> bm.intersection_update([8, 12, 55, 18])
        >>> bm
        BitMap([12])
        """
        for values in all_values:
            if isinstance(values, AbstractBitMap):
                self &= values
            else:
                self &= AbstractBitMap(values, copy_on_write=self.copy_on_write)

    def difference_update(self, *others):
        """
        Remove all elements of another set from this set.

        >>> bm = BitMap([1, 2, 3, 4, 5])
        >>> bm.difference_update(BitMap([1, 2, 10]), BitMap([3, 4, 20]))
        >>> bm
        BitMap([5])
        """
        self.__isub__(AbstractBitMap.union(*others))

    def symmetric_difference_update(self, other):
        """
        Update a set with the symmetric difference of itself and another.

        >>> bm = BitMap([1, 2, 3, 4])
        >>> bm.symmetric_difference_update(BitMap([1, 2, 10]))
        >>> bm
        BitMap([3, 4, 10])

        """
        self.__ixor__(other)
        
    def overwrite(self, AbstractBitMap other):
        """
        Clear the bitmap and overwrite it with another.

        >>> bm = BitMap([3, 12])
        >>> other = BitMap([4, 14])
        >>> bm.overwrite(other)
        >>> other.remove(4)
        >>> bm
        BitMap([4, 14])
        >>> other
        BitMap([14])
        """
        if self._c_bitmap == other._c_bitmap:
            raise ValueError('Cannot overwrite itself')
        croaring.roaring_bitmap_overwrite(self._c_bitmap, other._c_bitmap)

    def clear(self):
        """
        Remove all elements from this set.

        >>> bm = BitMap([1, 2, 3])
        >>> bm.clear()
        >>> bm
        BitMap([])
        """
        croaring.roaring_bitmap_clear(self._c_bitmap)

    def pop(self):
        """
        Remove and return an arbitrary set element.
        Raises KeyError if the set is empty.

        >>> bm = BitMap([1, 2])
        >>> a = bm.pop()
        >>> b = bm.pop()
        >>> bm
        BitMap([])
        >>> bm.pop()
        Traceback (most recent call last):
        ...
        KeyError: 'pop from an empty BitMap'

        """
        try:
            value = self.min()
        except ValueError:
            raise KeyError('pop from an empty BitMap')
        self.remove(value)
        return value


    def flip_inplace(self, uint64_t start, uint64_t end):
        """
        Compute (in place) the negation of the bitmap within the specified interval.

        Areas outside the range are passed unchanged.

        >>> bm = BitMap([3, 12])
        >>> bm.flip_inplace(10, 15)
        >>> bm
        BitMap([3, 10, 11, 13, 14])
        """
        croaring.roaring_bitmap_flip_inplace(self._c_bitmap, start, end)

    def add_range(self, uint64_t range_start, uint64_t range_end):
        """
        Add a range of values from range_start (included) to range_end (excluded).

        >>> bm = BitMap([5, 7])
        >>> bm.add_range(6, 9)
        >>> bm
        BitMap([5, 6, 7, 8])
        """
        if range_end <= range_start or range_end == 0 or range_start >= 2**32:
            return
        if range_end >= 2**32:
            range_end = 2**32
        croaring.roaring_bitmap_add_range(self._c_bitmap, range_start, range_end)

    def remove_range(self, uint64_t range_start, uint64_t range_end):
        """
        Remove a range of values from range_start (included) to range_end (excluded).

        >>> bm = BitMap([5, 6, 7, 8, 9, 10])
        >>> bm.remove_range(6, 9)
        >>> bm
        BitMap([5, 9, 10])
        """
        if range_end <= range_start or range_end == 0 or range_start >= 2**32:
            return
        if range_end >= 2**32:
            range_end = 2**32
        croaring.roaring_bitmap_remove_range(self._c_bitmap, range_start, range_end)

cdef class BitMap64(AbstractBitMap64):

    cdef compute_hash(self):
        '''Unsupported method.'''
        # For some reason, if we directly override __hash__ (either in BitMap or in FrozenBitMap), the __richcmp__
        # method disappears.
        raise TypeError('Cannot compute the hash of a %s.' % self.__class__.__name__)

    def add(self, uint64_t value):
        """
        Add an element to the bitmap. This has no effect if the element is already present.

        >>> bm = BitMap64()
        >>> bm.add(42)
        >>> bm
        BitMap64([42])
        >>> bm.add(42)
        >>> bm
        BitMap64([42])
        """
        croaring.roaring64_bitmap_add(self._c_bitmap, value)

    def add_checked(self, uint64_t value):
        """
        Add an element to the bitmap. This raises a KeyError exception if the element is already present.

        >>> bm = BitMap64()
        >>> bm.add_checked(42)
        >>> bm
        BitMap64([42])
        >>> bm.add_checked(42)
        Traceback (most recent call last):
        ...
        KeyError: 42
        """
        cdef bool test = croaring.roaring64_bitmap_add_checked(self._c_bitmap, value)
        if not test:
            raise KeyError(value)

    def update(self, *all_values): # FIXME could be more efficient
        """
        Add all the given values to the bitmap.

        >>> bm = BitMap64([3, 12])
        >>> bm.update([8, 12, 55, 18])
        >>> bm
        BitMap64([3, 8, 12, 18, 55])
        """
        cdef vector[uint64_t] buff_vect
        cdef uint64_t[:] buff
        for values in all_values:
            if isinstance(values, AbstractBitMap64):
                self |= values
            elif isinstance(values, range):
                if len(values) == 0:
                    continue
                _, (start, stop, step) = values.__reduce__()
                if step == -1:
                    step = 1
                    start, stop = stop+1, start+1
                if step == 1:
                    self.add_range(start, stop)
                else:
                    self |= AbstractBitMap64(values)
            elif isinstance(values, array.array) and len(values) > 0:
                buff = <array.array> values
                croaring.roaring64_bitmap_add_many(self._c_bitmap, len(values), &buff[0])
            else:
                try:
                    size = len(values)
                except TypeError:  # object has no length, creating a list
                    values = list(values)
                    size = len(values)
                if size > 0:
                    buff_vect = values
                    croaring.roaring64_bitmap_add_many(self._c_bitmap, size, &buff_vect[0])

    def discard(self, uint64_t value):
        """
        Remove an element from the bitmap. This has no effect if the element is not present.

        >>> bm = BitMap64([3, 12])
        >>> bm.discard(3)
        >>> bm
        BitMap64([12])
        >>> bm.discard(3)
        >>> bm
        BitMap64([12])
        """
        croaring.roaring64_bitmap_remove(self._c_bitmap, value)

    def remove(self, uint64_t value):
        """
        Remove an element from the bitmap. This raises a KeyError exception if the element does not exist in the bitmap.

        >>> bm = BitMap64([3, 12])
        >>> bm.remove(3)
        >>> bm
        BitMap64([12])
        >>> bm.remove(3)
        Traceback (most recent call last):
        ...
        KeyError: 3
        """
        cdef bool test = croaring.roaring64_bitmap_remove_checked(self._c_bitmap, value)
        if not test:
            raise KeyError(value)

    cdef binary_iop(self, AbstractBitMap64 other, (void)func(croaring.roaring64_bitmap_t*, const croaring.roaring64_bitmap_t*) noexcept) noexcept:
        func(self._c_bitmap, other._c_bitmap)
        return self

    def __ior__(self, other):
        if self._c_bitmap == (<AbstractBitMap64?>other)._c_bitmap:
            return self
        return (<BitMap64>self).binary_iop(<AbstractBitMap64?>other, croaring.roaring64_bitmap_or_inplace)

    def __iand__(self, other):
        if self._c_bitmap == (<AbstractBitMap64?>other)._c_bitmap:
            return self
        return (<BitMap64>self).binary_iop(<AbstractBitMap64?>other, croaring.roaring64_bitmap_and_inplace)

    def __ixor__(self, other):
        if self._c_bitmap == (<AbstractBitMap64?>other)._c_bitmap:
            self.clear()
            return self
        return (<BitMap64>self).binary_iop(<AbstractBitMap64?>other, croaring.roaring64_bitmap_xor_inplace)

    def __isub__(self, other):
        if self._c_bitmap == (<AbstractBitMap64?>other)._c_bitmap:
            self.clear()
            return self
        return (<BitMap64>self).binary_iop(<AbstractBitMap64?>other, croaring.roaring64_bitmap_andnot_inplace)

    def intersection_update(self, *all_values): # FIXME could be more efficient
        """
        Update the bitmap by taking its intersection with the given values.

        >>> bm = BitMap64([3, 12])
        >>> bm.intersection_update([8, 12, 55, 18])
        >>> bm
        BitMap64([12])
        """
        for values in all_values:
            if isinstance(values, AbstractBitMap64):
                self &= values
            else:
                self &= AbstractBitMap64(values)

    def difference_update(self, *others):
        """
        Remove all elements of another set from this set.

        >>> bm = BitMap64([1, 2, 3, 4, 5])
        >>> bm.difference_update(BitMap64([1, 2, 10]), BitMap64([3, 4, 20]))
        >>> bm
        BitMap64([5])
        """
        self.__isub__(AbstractBitMap64.union(*others))

    def symmetric_difference_update(self, other):
        """
        Update a set with the symmetric difference of itself and another.

        >>> bm = BitMap64([1, 2, 3, 4])
        >>> bm.symmetric_difference_update(BitMap64([1, 2, 10]))
        >>> bm
        BitMap64([3, 4, 10])

        """
        self.__ixor__(other)

    def overwrite(self, AbstractBitMap64 other):
        """
        Clear the bitmap and overwrite it with another.

        >>> bm = BitMap64([3, 12])
        >>> other = BitMap64([4, 14])
        >>> bm.overwrite(other)
        >>> other.remove(4)
        >>> bm
        BitMap64([4, 14])
        >>> other
        BitMap64([14])
        """
        if self._c_bitmap == other._c_bitmap:
            raise ValueError('Cannot overwrite itself')
        croaring.roaring64_bitmap_overwrite(self._c_bitmap, other._c_bitmap)

    def clear(self):
        """
        Remove all elements from this set.

        >>> bm = BitMap64([1, 2, 3])
        >>> bm.clear()
        >>> bm
        BitMap64([])
        """
        self.__iand__(BitMap64())

    def pop(self):
        """
        Remove and return an arbitrary set element.
        Raises KeyError if the set is empty.

        >>> bm = BitMap64([1, 2])
        >>> a = bm.pop()
        >>> b = bm.pop()
        >>> bm
        BitMap64([])
        >>> bm.pop()
        Traceback (most recent call last):
        ...
        KeyError: 'pop from an empty BitMap64'

        """
        try:
            value = self.min()
        except ValueError:
            raise KeyError('pop from an empty BitMap64')
        self.remove(value)
        return value

    def add_range(self, uint64_t range_start, uint64_t range_end):
        """
        Add a range of values from range_start (included) to range_end (excluded).

        >>> bm = BitMap64([5, 7])
        >>> bm.add_range(6, 9)
        >>> bm
        BitMap64([5, 6, 7, 8])
        """
        if range_end <= range_start or range_end == 0:
            return
        croaring.roaring64_bitmap_add_range(self._c_bitmap, range_start, range_end)

    def flip_inplace(self, uint64_t start, uint64_t end):
        """
        Compute (in place) the negation of the bitmap within the specified interval.

        Areas outside the range are passed unchanged.

        >>> bm = BitMap64([3, 12])
        >>> bm.flip_inplace(10, 15)
        >>> bm
        BitMap64([3, 10, 11, 13, 14])
        """
        croaring.roaring64_bitmap_flip_inplace(self._c_bitmap, start, end)

    def remove_range(self, uint64_t range_start, uint64_t range_end):
        """
        Remove a range of values from range_start (included) to range_end (excluded).

        >>> bm = BitMap64([5, 6, 7, 8, 9, 10])
        >>> bm.remove_range(6, 9)
        >>> bm
        BitMap64([5, 9, 10])
        """
        if range_end <= range_start or range_end == 0:
            return
        croaring.roaring64_bitmap_remove_range(self._c_bitmap, range_start, range_end)


#! /usr/bin/env python3

import sys
import random
import timeit

from pandas import Series, DataFrame

try:
    import tabulate
    has_tabulate = True
except ImportError:
    has_tabulate = False
    sys.stderr.write('Warning: could not import tabulate\n')
    sys.stderr.write('         see https://bitbucket.org/astanin/python-tabulate\n')
from pyroaring import BitMap, BitMap64

classes = {'set': set, 'pyroaring (32 bits)': BitMap, 'pyroaring (64 bits)': BitMap64, }
nb_exp = 30
size = int(1e6)
density = 0.125
universe_size = int(size / density)

try:
    from roaringbitmap import RoaringBitmap
    classes['roaringbitmap'] = RoaringBitmap
except ImportError:
    sys.stderr.write('Warning: could not import roaringbitmap\n')
    sys.stderr.write('         see https://github.com/andreasvc/roaringbitmap/\n')

try:
    from sortedcontainers.sortedset import SortedSet
    classes['sortedcontainers'] = SortedSet
except ImportError:
    sys.stderr.write('Warning: could not import sortedcontainers\n')
    sys.stderr.write('         see https://github.com/grantjenks/sorted_containers\n')

try:
    from croaring import BitSet
    classes['python-croaring'] = BitSet
except ImportError:
    sys.stderr.write('Warning: could not import croaring\n')
    sys.stderr.write('         see https://github.com/sunzhaoping/python-croaring\n')

import_str = 'import array, pickle; from __main__ import %s' % (','.join(
    ['get_list', 'get_range', 'random', 'size', 'universe_size']
    + [cls.__name__ for cls in classes.values() if cls is not set]))


def run_exp(stmt, setup, number):
    setup = '%s ; %s' % (import_str, setup)
    try:
        return timeit.timeit(stmt=stmt, setup=setup, number=number) / number
    except Exception:
        return float('nan')


def get_range():
    r = (0, universe_size, int(1 / density))
    try:
        return xrange(*r)
    except NameError:
        return range(*r)


def get_list():
    return random.sample(range(universe_size), size)


constructor = 'x={class_name}(values)'
simple_setup_constructor = 'x={class_name}(get_list());val=random.randint(0, universe_size)'
double_setup_constructor = 'x={class_name}(get_list()); y={class_name}(get_list())'
equal_setup_constructor = 'l=get_list();x={class_name}(l); y={class_name}(l)'
experiments = [
    # Constructors
    ('range constructor', ('values=get_range()', constructor)),
    ('ordered list constructor', ('values=get_list(); values.sort()', constructor)),
    ('list constructor', ('values=get_list()', constructor)),
    ('ordered array constructor', ('l=get_list(); l.sort(); values=array.array("I", l)', constructor)),
    ('array constructor', ('values=array.array("I", get_list())', constructor)),
    # Simple operations
    ('element addition', (simple_setup_constructor, 'x.add(val)')),
    ('element removal', (simple_setup_constructor, 'x.discard(val)')),
    ('membership test', (simple_setup_constructor, 'val in x')),
    # Binary operations
    ('union', (double_setup_constructor, 'z=x|y')),
    ('intersection', (double_setup_constructor, 'z=x&y')),
    ('difference', (double_setup_constructor, 'z=x-y')),
    ('symmetric diference', (double_setup_constructor, 'z=x^y')),
    ('equality test', (equal_setup_constructor, 'x==y')),
    ('subset test', (equal_setup_constructor, 'x<=y')),
    # Export
    ('conversion to list', (simple_setup_constructor, 'list(x)')),
    ('pickle dump & load', (simple_setup_constructor, 'pickle.loads(pickle.dumps(x, protocol=pickle.HIGHEST_PROTOCOL))')),
    ('"naive" conversion to array', (simple_setup_constructor, 'array.array("I", x)')),
    ('"optimized" conversion to array', (simple_setup_constructor, 'x.to_array()')),
    # Items
    ('selection', (simple_setup_constructor, 'x[int(size/2)]')),
    ('contiguous slice', (simple_setup_constructor, 'x[int(size/4):int(3*size/4):1]')),
    ('slice', (simple_setup_constructor, 'x[int(size/4):int(3*size/4):3]')),
    ('small slice', (simple_setup_constructor, 'x[int(size/100):int(3*size/100):3]')),
]
exp_dict = dict(experiments)


def run(cls, op):
    cls_name = classes[cls].__name__
    setup = exp_dict[op][0].format(class_name=cls_name)
    stmt = exp_dict[op][1].format(class_name=cls_name)
    result = run_exp(stmt=stmt, setup=setup, number=nb_exp)
    return result


def run_all():
    all_results = []
    for op, _ in experiments:
        sys.stderr.write('experiment: %s\n' % op)
        result = {'operation': op}
        for cls in random.sample(list(classes), len(classes)):
            result[cls] = run(cls, op)
        all_results.append(result)
    return DataFrame(all_results).sort_index(axis=1)


if __name__ == '__main__':
    df = run_all()
    print()
    if has_tabulate:
        print(tabulate.tabulate(df, headers='keys', tablefmt='rst', showindex='never', floatfmt=".2e"))
    else:
        print(df)



"""
Helper script to download a specific release amalgamation file for CRoaring.

Usage: python download_amalgamation.py <croaring_release_version, like v0.6.0>

The version needs to be the specific release tag on github.

"""
import os
import sys
from urllib.request import urlretrieve

version = sys.argv[1]

release = f"https://github.com/RoaringBitmap/CRoaring/releases/download/{version}/"

print(f"Downloading version {version} of the croaring amalgamation")

files = ["roaring.c", "roaring.h"]

for file in files:
    r = urlretrieve(release + file, os.path.join("pyroaring", file))

with open(os.path.join("pyroaring", "croaring_version.pxi"), "w") as f:
    f.write(f"__croaring_version__ = \"{version}\"")



# distutils: language = c++
# cython: binding = True, language_level=3

include 'version.pxi'
include 'croaring_version.pxi'
include 'abstract_bitmap.pxi'
include 'frozen_bitmap.pxi'
include 'bitmap.pxi'



from libc.stdint cimport uint8_t, int32_t, uint32_t, uint64_t, int64_t
from libcpp cimport bool

cdef extern from "roaring.h":
    ctypedef struct roaring_array_t:
        pass
    ctypedef struct roaring_bitmap_t:
        roaring_array_t high_low_container
    ctypedef struct roaring_uint32_iterator_t:
        const roaring_bitmap_t *parent
        int32_t container_index
        int32_t in_container_index
        int32_t run_index
        uint32_t in_run_index
        uint32_t current_value
        bool has_value
        const void *container
        uint8_t typecode
        uint32_t highbits
    ctypedef struct roaring_statistics_t:
        uint32_t n_containers
        uint32_t n_array_containers
        uint32_t n_run_containers
        uint32_t n_bitset_containers
        uint32_t n_values_array_containers
        uint32_t n_values_run_containers
        uint32_t n_values_bitset_containers
        uint32_t n_bytes_array_containers
        uint32_t n_bytes_run_containers
        uint32_t n_bytes_bitset_containers
        uint32_t max_value
        uint32_t min_value
        uint64_t sum_value
        uint64_t cardinality
    ctypedef struct roaring64_statistics_t:
        uint64_t n_containers
        uint64_t n_array_containers
        uint64_t n_run_containers
        uint64_t n_bitset_containers
        uint64_t n_values_array_containers
        uint64_t n_values_run_containers
        uint64_t n_values_bitset_containers
        uint64_t n_bytes_array_containers
        uint64_t n_bytes_run_containers
        uint64_t n_bytes_bitset_containers
        uint64_t max_value
        uint64_t min_value
        uint64_t cardinality

    roaring_bitmap_t *roaring_bitmap_create()
    bool roaring_bitmap_get_copy_on_write(const roaring_bitmap_t* r)
    void roaring_bitmap_set_copy_on_write(roaring_bitmap_t* r, bool cow)
    void roaring_bitmap_add(roaring_bitmap_t *r, uint32_t x)
    bool roaring_bitmap_add_checked(roaring_bitmap_t *r, uint32_t x)
    void roaring_bitmap_add_many(roaring_bitmap_t *r, size_t n_args, const uint32_t *vals)
    void roaring_bitmap_add_range(roaring_bitmap_t *ra, uint64_t min, uint64_t max);
    void roaring_bitmap_remove(roaring_bitmap_t *r, uint32_t x)
    void roaring_bitmap_remove_range(roaring_bitmap_t *ra, uint64_t min, uint64_t max)
    bool roaring_bitmap_remove_checked(roaring_bitmap_t *r, uint32_t x)
    void roaring_bitmap_clear(roaring_bitmap_t *r)
    bool roaring_bitmap_contains(const roaring_bitmap_t *r, uint32_t val)
    bool roaring_bitmap_contains_range(const roaring_bitmap_t *r, uint64_t range_start, uint64_t range_end)
    roaring_bitmap_t *roaring_bitmap_copy(const roaring_bitmap_t *r)
    bool roaring_bitmap_overwrite(roaring_bitmap_t *dest, const roaring_bitmap_t *src)
    roaring_bitmap_t *roaring_bitmap_from_range(uint64_t min, uint64_t max, uint32_t step)
    bool roaring_bitmap_run_optimize(roaring_bitmap_t *r)
    size_t roaring_bitmap_shrink_to_fit(roaring_bitmap_t *r)
    void roaring_bitmap_free(roaring_bitmap_t *r)
    roaring_bitmap_t *roaring_bitmap_of_ptr(size_t n_args, const uint32_t *vals)
    uint64_t roaring_bitmap_get_cardinality(const roaring_bitmap_t *r)
    uint64_t roaring_bitmap_range_cardinality(const roaring_bitmap_t *r, uint64_t range_start, uint64_t range_end)
    bool roaring_bitmap_is_empty(const roaring_bitmap_t *ra)
    bool roaring_bitmap_equals(const roaring_bitmap_t *r1, const roaring_bitmap_t *r2)
    bool roaring_bitmap_is_strict_subset(const roaring_bitmap_t *r1, const roaring_bitmap_t *r2)
    bool roaring_bitmap_is_subset(const roaring_bitmap_t *r1, const roaring_bitmap_t *r2)
    void roaring_bitmap_to_uint32_array(const roaring_bitmap_t *r, uint32_t *ans)
    roaring_bitmap_t *roaring_bitmap_or_many(size_t number, const roaring_bitmap_t **x)
    roaring_bitmap_t *roaring_bitmap_or(const roaring_bitmap_t *x1, const roaring_bitmap_t *x2)
    void roaring_bitmap_or_inplace(roaring_bitmap_t *x1, const roaring_bitmap_t *x2)
    roaring_bitmap_t *roaring_bitmap_and(const roaring_bitmap_t *x1, const roaring_bitmap_t *x2)
    void roaring_bitmap_and_inplace(roaring_bitmap_t *x1, const roaring_bitmap_t *x2)
    roaring_bitmap_t *roaring_bitmap_xor(const roaring_bitmap_t *x1, const roaring_bitmap_t *x2)
    void roaring_bitmap_xor_inplace(roaring_bitmap_t *x1, const roaring_bitmap_t *x2)
    roaring_bitmap_t *roaring_bitmap_andnot(const roaring_bitmap_t *x1, const roaring_bitmap_t *x2)
    void roaring_bitmap_andnot_inplace(roaring_bitmap_t *x1, const roaring_bitmap_t *x2)
    uint64_t roaring_bitmap_or_cardinality(const roaring_bitmap_t *x1, const roaring_bitmap_t *x2)
    uint64_t roaring_bitmap_and_cardinality(const roaring_bitmap_t *x1, const roaring_bitmap_t *x2)
    uint64_t roaring_bitmap_andnot_cardinality(const roaring_bitmap_t *x1, const roaring_bitmap_t *x2)
    uint64_t roaring_bitmap_xor_cardinality(const roaring_bitmap_t *x1, const roaring_bitmap_t *x2)
    bool roaring_bitmap_intersect(const roaring_bitmap_t *x1, const roaring_bitmap_t *x2)
    double roaring_bitmap_jaccard_index(const roaring_bitmap_t *x1, const roaring_bitmap_t *x2)
    uint32_t roaring_bitmap_minimum(const roaring_bitmap_t *r)
    uint32_t roaring_bitmap_maximum(const roaring_bitmap_t *r)
    uint64_t roaring_bitmap_rank(const roaring_bitmap_t *r, uint32_t x)
    roaring_bitmap_t *roaring_bitmap_flip(const roaring_bitmap_t *x1, uint64_t range_start, uint64_t range_end)
    void roaring_bitmap_flip_inplace(roaring_bitmap_t *x1, uint64_t range_start, uint64_t range_end)
    roaring_bitmap_t *roaring_bitmap_add_offset(const roaring_bitmap_t *bm, int64_t offset)
    bool roaring_bitmap_select(const roaring_bitmap_t *r, uint32_t rank, uint32_t *element)
    void roaring_bitmap_statistics(const roaring_bitmap_t *r, roaring_statistics_t *stat)
    size_t roaring_bitmap_portable_size_in_bytes(const roaring_bitmap_t *ra)
    size_t roaring_bitmap_portable_serialize(const roaring_bitmap_t *ra, char *buf)
    roaring_bitmap_t *roaring_bitmap_portable_deserialize(const char *buf)
    roaring_bitmap_t *roaring_bitmap_portable_deserialize_safe(const char *buf, size_t maxbytes)
    bool roaring_bitmap_internal_validate(const roaring_bitmap_t *r, const char **reason)
    roaring_uint32_iterator_t *roaring_iterator_create(const roaring_bitmap_t *ra)
    bool roaring_uint32_iterator_advance(roaring_uint32_iterator_t *it)
    uint32_t roaring_uint32_iterator_read(roaring_uint32_iterator_t *it, uint32_t* buf, uint32_t count)
    bool roaring_uint32_iterator_move_equalorlarger(roaring_uint32_iterator_t *it, uint32_t val)
    void roaring_uint32_iterator_free(roaring_uint32_iterator_t *it)

    # 64-bit roaring bitmaps
    ctypedef struct roaring64_bitmap_t:
        pass

    ctypedef struct roaring64_iterator_t:
        pass

    roaring64_bitmap_t *roaring64_bitmap_create()
    void roaring64_bitmap_free(roaring64_bitmap_t *r)
    roaring64_bitmap_t *roaring64_bitmap_copy(const roaring64_bitmap_t *r)
    roaring64_bitmap_t *roaring64_bitmap_of_ptr(size_t n_args, const uint64_t *vals)
    roaring64_bitmap_t *roaring64_bitmap_from_range(uint64_t min, uint64_t max, uint64_t step)
    void roaring64_bitmap_add(roaring64_bitmap_t *r, uint64_t val)
    bool roaring64_bitmap_add_checked(roaring64_bitmap_t *r, uint64_t val)
    void roaring64_bitmap_add_many(roaring64_bitmap_t *r, size_t n_args, const uint64_t *vals)
    void roaring64_bitmap_add_range(roaring64_bitmap_t *r, uint64_t min, uint64_t max)
    void roaring64_bitmap_remove(roaring64_bitmap_t *r, uint64_t val)
    bool roaring64_bitmap_remove_checked(roaring64_bitmap_t *r, uint64_t val)
    void roaring64_bitmap_remove_many(roaring64_bitmap_t *r, size_t n_args, const uint64_t *vals)
    void roaring64_bitmap_remove_range(roaring64_bitmap_t *r, uint64_t min, uint64_t max)
    bool roaring64_bitmap_contains(const roaring64_bitmap_t *r, uint64_t val)
    bool roaring64_bitmap_contains_range(const roaring64_bitmap_t *r, uint64_t min, uint64_t max)
    bool roaring64_bitmap_select(const roaring64_bitmap_t *r, uint64_t rank, uint64_t *element)
    void roaring64_bitmap_statistics(const roaring64_bitmap_t *r, roaring64_statistics_t *stat)
    uint64_t roaring64_bitmap_rank(const roaring64_bitmap_t *r, uint64_t val)
    roaring64_bitmap_t *roaring64_bitmap_flip(const roaring64_bitmap_t *r, uint64_t min, uint64_t max)
    void roaring64_bitmap_flip_inplace(roaring64_bitmap_t *r, uint64_t min, uint64_t max)
    bool roaring64_bitmap_get_index(const roaring64_bitmap_t *r, uint64_t val, uint64_t *out_index)
    uint64_t roaring64_bitmap_get_cardinality(const roaring64_bitmap_t *r)
    uint64_t roaring64_bitmap_range_cardinality(const roaring64_bitmap_t *r, uint64_t min, uint64_t max)
    bool roaring64_bitmap_is_empty(const roaring64_bitmap_t *r)
    uint64_t roaring64_bitmap_minimum(const roaring64_bitmap_t *r)
    uint64_t roaring64_bitmap_maximum(const roaring64_bitmap_t *r)
    bool roaring64_bitmap_run_optimize(roaring64_bitmap_t *r)
    size_t roaring64_bitmap_shrink_to_fit(roaring64_bitmap_t *r)
    size_t roaring64_bitmap_size_in_bytes(const roaring64_bitmap_t *r)
    bool roaring64_bitmap_equals(const roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    bool roaring64_bitmap_is_subset(const roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    bool roaring64_bitmap_is_strict_subset(const roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    void roaring64_bitmap_to_uint64_array(const roaring64_bitmap_t *r, uint64_t *out)
    roaring64_bitmap_t *roaring64_bitmap_and(const roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    uint64_t roaring64_bitmap_and_cardinality(const roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    void roaring64_bitmap_and_inplace(roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    bool roaring64_bitmap_intersect(const roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    double roaring64_bitmap_jaccard_index(const roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    roaring64_bitmap_t *roaring64_bitmap_or(const roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    uint64_t roaring64_bitmap_or_cardinality(const roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    void roaring64_bitmap_or_inplace(roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    roaring64_bitmap_t *roaring64_bitmap_xor(const roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    uint64_t roaring64_bitmap_xor_cardinality(const roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    void roaring64_bitmap_xor_inplace(roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    roaring64_bitmap_t *roaring64_bitmap_andnot(const roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    uint64_t roaring64_bitmap_andnot_cardinality(const roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    void roaring64_bitmap_andnot_inplace(roaring64_bitmap_t *r1, const roaring64_bitmap_t *r2)
    size_t roaring64_bitmap_portable_size_in_bytes(const roaring64_bitmap_t *r)
    size_t roaring64_bitmap_portable_serialize(const roaring64_bitmap_t *r, char *buf)
    size_t roaring64_bitmap_portable_deserialize_size(const char *buf, size_t maxbytes)
    roaring64_bitmap_t *roaring64_bitmap_portable_deserialize_safe(const char *buf, size_t maxbytes)
    bool roaring64_bitmap_internal_validate(const roaring64_bitmap_t *r, const char **reason)
    roaring64_bitmap_t *roaring64_bitmap_add_offset_signed(const roaring64_bitmap_t *r, bool positive, uint64_t offset)
    void roaring64_bitmap_overwrite(roaring64_bitmap_t *dest, const roaring64_bitmap_t *src)
    roaring64_iterator_t *roaring64_iterator_create(const roaring64_bitmap_t *r)
    void roaring64_iterator_free(roaring64_iterator_t *it)
    bool roaring64_iterator_has_value(const roaring64_iterator_t *it)
    bool roaring64_iterator_advance(roaring64_iterator_t *it)
    uint64_t roaring64_iterator_value(const roaring64_iterator_t *it)
    bool roaring64_iterator_move_equalorlarger(roaring64_iterator_t *it, uint64_t val)
    uint64_t roaring64_iterator_read(roaring64_iterator_t *it, uint64_t *buf, uint64_t count)


#! /usr/bin/env python3

import os
import platform
from distutils.sysconfig import get_config_vars

from setuptools import setup
from setuptools.extension import Extension

PKG_DIR = 'pyroaring'

PLATFORM_WINDOWS = (platform.system() == 'Windows')
PLATFORM_MACOSX = (platform.system() == 'Darwin')

# Read version file from the src
with open("pyroaring/version.pxi") as fp:
    exec(fp.read())
    VERSION = __version__  # noqa: F821


# Remove -Wstrict-prototypes option
# See http://stackoverflow.com/a/29634231/4110059
if not PLATFORM_WINDOWS:
    cfg_vars = get_config_vars()
    for key, value in cfg_vars.items():
        if type(value) is str:
            cfg_vars[key] = value.replace("-Wstrict-prototypes", "")

try:
    with open('README.rst') as f:
        long_description = ''.join(f.readlines())
except (IOError, ImportError, RuntimeError):
    print('Could not generate long description.')
    long_description = ''


if PLATFORM_WINDOWS:
    # MSVC defaults to C++14, so no explicit C++ standard flag is needed.
    # roaring.c requires C11 for features like alignof and _Static_assert.
    ext_compile_args = []
    c_compile_args = ['/std:c11']
else:
    c_compile_args = ['-D__STDC_LIMIT_MACROS', '-D__STDC_CONSTANT_MACROS', '-D _GLIBCXX_ASSERTIONS']
    if PLATFORM_MACOSX:
        c_compile_args.append('-mmacosx-version-min=10.14')
    if 'DEBUG' in os.environ:
        c_compile_args.extend(['-O0', '-g'])
    else:
        c_compile_args.append('-O3')
    if 'ARCHI' in os.environ:
        if os.environ['ARCHI'] != "generic":
            c_compile_args.append('-march=%s' % os.environ['ARCHI'])
    # The '-march=native' flag is not universally allowed. In particular, it
    # will systematically fail on aarch64 systems (like the new Apple M1 systems). It
    # also creates troubles under macOS with pip installs and requires ugly workarounds.
    # The best way to handle people who want to use -march=native is to ask them
    # to pass ARCHI=native to their build process.
    # else:
    #    c_compile_args.append('-march=native')
    ext_compile_args = c_compile_args + ['-std=c++11']
    c_compile_args = c_compile_args + ['-std=c11']

pyroaring_module = Extension(
    'pyroaring',
    sources=[os.path.join(PKG_DIR, 'pyroaring.pyx')],
    extra_compile_args=ext_compile_args,
    language='c++',
)

# roaring.c is compiled as a separate static library because it must be compiled
# as C (not C++), with a C11 standard flag.
libraries = [(
    'croaring',
    {
        'sources': [os.path.join(PKG_DIR, 'roaring.c')],
        'cflags': c_compile_args,
    },
)]

setup(
    name='pyroaring',
    ext_modules=[pyroaring_module],
    libraries=libraries,
    package_data={'pyroaring': ['py.typed', '__init__.pyi']},
    packages=['pyroaring'],
    version=VERSION,
    description='Library for handling efficiently sorted integer sets.',
    long_description=long_description,
    setup_requires=['cython>=3.0.2'],
    url='https://github.com/Ezibenroc/PyRoaringBitMap',
    author='Tom Cornebize',
    author_email='tom.cornebize@gmail.com',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
    ],
)



import array
from typing import overload, TypedDict
from collections.abc import Iterable, Iterator

from typing_extensions import Self

__version__: str
__croaring_version__: str


class _Statistics(TypedDict):
    # Type as observed in the returned values.
    # Warning: This type does not exist at runtime.

    n_containers: int
    n_array_containers: int
    n_run_containers: int
    n_bitset_containers: int
    n_values_array_containers: int
    n_values_run_containers: int
    n_values_bitset_containers: int
    n_bytes_array_containers: int
    n_bytes_run_containers: int
    n_bytes_bitset_containers: int
    max_value: int
    min_value: int
    sum_value: int
    cardinality: int


class AbstractBitMap:
    def __init__(self, values: Iterable[int] | None = None, copy_on_write: bool = False, optimize: bool = True) -> None:
        ...

    @property
    def raw_pointer(self) -> int:
        ...

    @property
    def copy_on_write(self) -> bool:
        ...

    def run_optimize(self) -> bool:
        ...

    def shrink_to_fit(self) -> int:
        ...

    def __contains__(self, value: int) -> bool:
        ...

    def __bool__(self) -> bool:
        ...

    def __len__(self) -> int:
        ...

    def __lt__(self, other: AbstractBitMap) -> bool:
        ...

    def __le__(self, other: AbstractBitMap) -> bool:
        ...

    def __eq__(self, other: object) -> bool:
        ...

    def __ne__(self, other: object) -> bool:
        ...

    def __gt__(self, other: AbstractBitMap) -> bool:
        ...

    def __ge__(self, other: AbstractBitMap) -> bool:
        ...

    def contains_range(self, range_start: int, range_end: int) -> bool:
        ...

    def range_cardinality(self, range_start: int, range_end: int) -> int:
        ...

    def iter_equal_or_larger(self, val: int) -> Iterator[int]:
        ...

    def __iter__(self) -> Iterator[int]:
        ...

    def flip(self, start: int, end: int) -> Self:
        ...

    def shift(self, offset: int) -> Self:
        ...

    def copy(self) -> Self:
        ...

    def isdisjoint(self, other: AbstractBitMap) -> bool:
        ...

    def issubset(self, other: AbstractBitMap) -> bool:
        ...

    def issuperset(self, other: AbstractBitMap) -> bool:
        ...

    # Note: `difference` and others are sort-of set up like they're meant to be
    # static methods (accepting _only_ `*bitmaps` in the underlying Cython
    # code), however at runtime they require at least one argument and return an
    # instance of the same type as that value -- like instance methods. Typing
    # them as instances methods ensures that mypy matches this behaviour (other
    # type checkers untested), even when used statically as their docstrings
    # suggest.

    def difference(self, *bitmaps: AbstractBitMap) -> Self:
        ...

    def symmetric_difference(self, other: AbstractBitMap) -> Self:
        ...

    def union(self, *bitmaps: AbstractBitMap) -> Self:
        ...

    def intersection(self, *bitmaps: AbstractBitMap) -> Self:
        ...

    def __or__(self, other: AbstractBitMap) -> Self:
        ...

    def __and__(self, other: AbstractBitMap) -> Self:
        ...

    def __xor__(self, other: AbstractBitMap) -> Self:
        ...

    def __sub__(self, other: AbstractBitMap) -> Self:
        ...

    def union_cardinality(self, other: AbstractBitMap) -> int:
        ...

    def intersection_cardinality(self, other: AbstractBitMap) -> int:
        ...

    def difference_cardinality(self, other: AbstractBitMap) -> int:
        ...

    def symmetric_difference_cardinality(self, other: AbstractBitMap) -> int:
        ...

    def intersect(self, other: AbstractBitMap) -> bool:
        ...

    def jaccard_index(self, other: AbstractBitMap) -> float:
        ...

    def get_statistics(self) -> _Statistics:
        ...

    def min(self) -> int:
        ...

    def max(self) -> int:
        ...

    def rank(self, value: int) -> int:
        ...

    def next_set_bit(self, value: int) -> int:
        ...

    @overload
    def __getitem__(self, value: int) -> int:
        ...

    @overload
    def __getitem__(self, value: slice) -> Self:
        ...

    def serialize(self) -> bytes:
        ...

    @classmethod
    def deserialize(cls, buff: bytes) -> Self:
        ...

    def __getstate__(self) -> bytes:
        ...

    def __setstate__(self, state: bytes) -> Self:
        ...

    def __sizeof__(self) -> int:
        ...

    def to_array(self) -> array.array[int]:
        ...


class FrozenBitMap(AbstractBitMap):
    def __hash__(self) -> int:
        ...


class BitMap(AbstractBitMap):
    def add(self, value: int) -> None:
        ...

    def add_checked(self, value: int) -> None:
        ...

    def update(self, *all_values: Iterable[int]) -> None:
        ...

    def discard(self, value: int) -> None:
        ...

    def remove(self, value: int) -> None:
        ...

    def __ior__(self, other: AbstractBitMap) -> Self:
        ...

    def __iand__(self, other: AbstractBitMap) -> Self:
        ...

    def __ixor__(self, other: AbstractBitMap) -> Self:
        ...

    def __isub__(self, other: AbstractBitMap) -> Self:
        ...

    def intersection_update(self, *all_values: Iterable[int]) -> None:
        ...

    def difference_update(self, *others: AbstractBitMap) -> None:
        ...

    def symmetric_difference_update(self, other: AbstractBitMap) -> None:
        ...

    def overwrite(self, other: AbstractBitMap) -> None:
        ...

    def clear(self) -> None:
        ...

    def pop(self) -> int:
        ...

    def flip_inplace(self, start: int, end: int) -> None:
        ...

    def add_range(self, range_start: int, range_end: int) -> None:
        ...

    def remove_range(self, range_start: int, range_end: int) -> None:
        ...

class AbstractBitMap64:
    def __init__(self, values: Iterable[int] | None = None, copy_on_write: bool = False, optimize: bool = True) -> None:
        ...

    @property
    def raw_pointer(self) -> int:
        ...

    @property
    def copy_on_write(self) -> bool:
        ...

    def run_optimize(self) -> bool:
        ...

    def shrink_to_fit(self) -> int:
        ...

    def __contains__(self, value: int) -> bool:
        ...

    def __bool__(self) -> bool:
        ...

    def __len__(self) -> int:
        ...

    def __lt__(self, other: AbstractBitMap64) -> bool:
        ...

    def __le__(self, other: AbstractBitMap64) -> bool:
        ...

    def __eq__(self, other: object) -> bool:
        ...

    def __ne__(self, other: object) -> bool:
        ...

    def __gt__(self, other: AbstractBitMap64) -> bool:
        ...

    def __ge__(self, other: AbstractBitMap64) -> bool:
        ...

    def contains_range(self, range_start: int, range_end: int) -> bool:
        ...

    def range_cardinality(self, range_start: int, range_end: int) -> int:
        ...

    def iter_equal_or_larger(self, val: int) -> Iterator[int]:
        ...

    def __iter__(self) -> Iterator[int]:
        ...

    def flip(self, start: int, end: int) -> Self:
        ...

    def shift(self, offset: int) -> Self:
        ...

    def copy(self) -> Self:
        ...

    def isdisjoint(self, other: AbstractBitMap64) -> bool:
        ...

    def issubset(self, other: AbstractBitMap64) -> bool:
        ...

    def issuperset(self, other: AbstractBitMap64) -> bool:
        ...

    def difference(self, *bitmaps: AbstractBitMap64) -> Self:
        ...

    def symmetric_difference(self, other: AbstractBitMap64) -> Self:
        ...

    def union(self, *bitmaps: AbstractBitMap64) -> Self:
        ...

    def intersection(self, *bitmaps: AbstractBitMap64) -> Self:
        ...

    def __or__(self, other: AbstractBitMap64) -> Self:
        ...

    def __and__(self, other: AbstractBitMap64) -> Self:
        ...

    def __xor__(self, other: AbstractBitMap64) -> Self:
        ...

    def __sub__(self, other: AbstractBitMap64) -> Self:
        ...

    def union_cardinality(self, other: AbstractBitMap64) -> int:
        ...

    def intersection_cardinality(self, other: AbstractBitMap64) -> int:
        ...

    def difference_cardinality(self, other: AbstractBitMap64) -> int:
        ...

    def symmetric_difference_cardinality(self, other: AbstractBitMap64) -> int:
        ...

    def intersect(self, other: AbstractBitMap64) -> bool:
        ...

    def jaccard_index(self, other: AbstractBitMap64) -> float:
        ...

    def get_statistics(self) -> _Statistics:
        ...

    def min(self) -> int:
        ...

    def max(self) -> int:
        ...

    def rank(self, value: int) -> int:
        ...

    def next_set_bit(self, value: int) -> int:
        ...

    @overload
    def __getitem__(self, value: int) -> int:
        ...

    @overload
    def __getitem__(self, value: slice) -> Self:
        ...

    def serialize(self) -> bytes:
        ...

    @classmethod
    def deserialize(cls, buff: bytes) -> Self:
        ...

    def __getstate__(self) -> bytes:
        ...

    def __setstate__(self, state: bytes) -> Self:
        ...

    def __sizeof__(self) -> int:
        ...

    def to_array(self) -> array.array[int]:
        ...


class FrozenBitMap64(AbstractBitMap64):
    def __hash__(self) -> int:
        ...


class BitMap64(AbstractBitMap64):
    def add(self, value: int) -> None:
        ...

    def add_checked(self, value: int) -> None:
        ...

    def update(self, *all_values: Iterable[int]) -> None:
        ...

    def discard(self, value: int) -> None:
        ...

    def remove(self, value: int) -> None:
        ...

    def __ior__(self, other: AbstractBitMap64) -> Self:
        ...

    def __iand__(self, other: AbstractBitMap64) -> Self:
        ...

    def __ixor__(self, other: AbstractBitMap64) -> Self:
        ...

    def __isub__(self, other: AbstractBitMap64) -> Self:
        ...

    def intersection_update(self, *all_values: Iterable[int]) -> None:
        ...

    def difference_update(self, *others: AbstractBitMap64) -> None:
        ...

    def symmetric_difference_update(self, other: AbstractBitMap64) -> None:
        ...

    def overwrite(self, other: AbstractBitMap64) -> None:
        ...

    def clear(self) -> None:
        ...

    def pop(self) -> int:
        ...

    def flip_inplace(self, start: int, end: int) -> None:
        ...

    def add_range(self, range_start: int, range_end: int) -> None:
        ...

    def remove_range(self, range_start: int, range_end: int) -> None:
        ...